#
#  tables.py
#

import json
import types
from collections import defaultdict
from os.path import dirname, join, splitext
from pathlib import Path
from typing import (
    IO,
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

import numpy as np
import pandas as pd
import pyarrow
import pyarrow.parquet as pq
import structlog
from owid.repack import repack_frame
from pandas._typing import Scalar
from pandas.core.series import Series
from pandas.util._decorators import rewrite_axis_style_signature

from . import variables
from .meta import License, Origin, Source, TableMeta, VariableMeta
from .utils import underscore

log = structlog.get_logger()

SCHEMA = json.load(open(join(dirname(__file__), "schemas", "table.json")))
METADATA_FIELDS = list(SCHEMA["properties"])

# New type required for pandas reading functions.
AnyStr = TypeVar("AnyStr", str, bytes)


class Table(pd.DataFrame):
    # metdata about the entire table
    metadata: TableMeta

    # metadata about individual columns
    # NOTE: the name _fields is also on the Variable class, pandas will propagate this to
    #       any slices, which is how they get access to their metadata
    _fields: Dict[str, VariableMeta]

    # propagate all these fields on every slice or copy
    _metadata = ["metadata", "_fields"]

    # slicing and copying creates tables
    @property
    def _constructor(self) -> type:
        return Table

    @property
    def _constructor_sliced(self) -> Any:
        return variables.Variable

    def __init__(
        self,
        *args: Any,
        metadata: Optional[TableMeta] = None,
        short_name: Optional[str] = None,
        underscore=False,
        camel_to_snake=False,
        like: Optional["Table"] = None,
        **kwargs: Any,
    ) -> None:
        """
        :param metadata: TableMeta to use
        :param short_name: Use empty TableMeta and fill it with `short_name`. This is a shorter version
            of `Table(df, metadata=TableMeta(short_name="my_name"))`
        :param underscore: Underscore table columns and indexes. See `underscore` method for help
        :param camel_to_snake: Convert camelCase column names to snake_case.
        :param like: Use metadata from Table given in this argument (including columns). This is a shorter version of
            new_t = Table(df, metadata=old_t.metadata)
            for col in new_t.columns:
                new_t[col].metadata = deepcopy(old_t[col].metadata)
        """

        super().__init__(*args, **kwargs)

        # empty table metadata by default
        self.metadata = metadata or TableMeta()

        # use supplied short_name
        if short_name:
            assert self.metadata.short_name is None or (
                self.metadata.short_name == short_name
            ), "short_name is different from the one in metadata"
            self.metadata.short_name = short_name

        # all columns have empty metadata by default
        assert not hasattr(self, "_fields")
        self._fields = defaultdict(VariableMeta)

        # underscore column names
        if underscore:
            self.underscore(inplace=True, camel_to_snake=camel_to_snake)

        # reuse metadata from a different table
        if like is not None:
            copy = self.copy_metadata(like)
            self._fields = copy._fields
            self.metadata = copy.metadata

    @property
    def m(self) -> TableMeta:
        """Metadata alias to save typing."""
        return self.metadata

    @property
    def primary_key(self) -> List[str]:
        return [n for n in self.index.names if n]

    def to(self, path: Union[str, Path], repack: bool = True) -> None:
        """
        Save this table in one of our SUPPORTED_FORMATS.
        """
        # Add entry in the processing log about operation "save".
        self = update_processing_logs_when_saving_table(table=self, path=path)

        if isinstance(path, Path):
            path = path.as_posix()

        if path.endswith(".csv"):
            # ignore repacking
            return self.to_csv(path)

        elif path.endswith(".feather"):
            return self.to_feather(path, repack=repack)

        elif path.endswith(".parquet"):
            return self.to_parquet(path, repack=repack)

        else:
            raise ValueError(f"could not detect a suitable format to save to: {path}")

    @classmethod
    def read(cls, path: Union[str, Path]) -> "Table":
        if isinstance(path, Path):
            path = path.as_posix()

        if path.endswith(".csv"):
            table = cls.read_csv(path)

        elif path.endswith(".feather"):
            table = cls.read_feather(path)

        elif path.endswith(".parquet"):
            table = cls.read_parquet(path)
        else:
            raise ValueError(f"could not detect a suitable format to read from: {path}")

        # Add processing log to the metadata of each variable in the table.
        # TODO: For some reason, the snapshot loading entry gets repeated.
        table = update_processing_logs_when_loading_or_creating_table(table=table)

        return table

    # Mypy complaints about this not matching the defintiion of NDFrame.to_csv but I don't understand why
    def to_csv(self, path: Any, **kwargs: Any) -> None:  # type: ignore
        """
        Save this table as a csv file plus accompanying JSON metadata file.
        If the table is stored at "mytable.csv", the metadata will be at
        "mytable.meta.json".
        """
        if not str(path).endswith(".csv"):
            raise ValueError(f'filename must end in ".csv": {path}')

        df = pd.DataFrame(self)
        # if the dataframe uses the default index then we don't want to store it (would be a column of row numbers)
        save_index = self.primary_key != []
        df.to_csv(path, index=save_index, **kwargs)

        metadata_filename = splitext(path)[0] + ".meta.json"
        self._save_metadata(metadata_filename)

    def to_feather(
        self,
        path: Any,
        repack: bool = True,
        compression: Literal["zstd", "lz4", "uncompressed"] = "zstd",
        **kwargs: Any,
    ) -> None:
        """
        Save this table as a feather file plus accompanying JSON metadata file.
        If the table is stored at "mytable.feather", the metadata will be at
        "mytable.meta.json".
        """
        if not str(path).endswith(".feather"):
            raise ValueError(f'filename must end in ".feather": {path}')

        # feather can't store the index
        df = pd.DataFrame(self)
        if self.primary_key:
            overlapping_names = set(self.index.names) & set(self.columns)
            if overlapping_names:
                raise ValueError(f"index names are overlapping with column names: {overlapping_names}")
            df = df.reset_index()

        if repack:
            # use smaller data types wherever possible
            # NOTE: this can be slow for large dataframes
            df = repack_frame(df)

        df.to_feather(path, compression=compression, **kwargs)

        self._save_metadata(self.metadata_filename(path))

    def metadata_filename(self, path: str):
        return splitext(path)[0] + ".meta.json"

    def to_parquet(self, path: Any, repack: bool = True) -> None:  # type: ignore
        """
        Save this table as a parquet file with embedded metadata in the table schema.

        NOTE: we save the metadata for fields in the table scheme, but it might be
              possible with Parquet to store it in the fields themselves somehow
        """
        if not str(path).endswith(".parquet"):
            raise ValueError(f'filename must end in ".parquet": {path}')

        # parquet can store the index, but repacking is wasted on index columns so
        # we get rid of the index first
        df = pd.DataFrame(self)
        if self.primary_key:
            df = df.reset_index()

        if repack:
            # use smaller data types wherever possible
            # NOTE: this can be slow for large dataframes
            df = repack_frame(df)

        # create a pyarrow table with metadata in the schema
        # (some metadata gets auto-generated to help pandas deserialise better, we want to keep that)
        t = pyarrow.Table.from_pandas(df)

        # adding metadata would make reading partial content inefficient, see https://github.com/owid/etl/issues/783
        # new_metadata = {
        #     b"owid_table": json.dumps(self.metadata.to_dict(), default=str),  # type: ignore
        #     b"owid_fields": json.dumps(self._get_fields_as_dict(), default=str),
        #     b"primary_key": json.dumps(self.primary_key),
        #     **t.schema.metadata,
        # }
        # schema = t.schema.with_metadata(new_metadata)
        # t = t.cast(schema)

        # write the combined table to disk
        pq.write_table(t, path)

        self._save_metadata(self.metadata_filename(path))

    def _save_metadata(self, filename: str) -> None:
        # write metadata
        with open(filename, "w") as ostream:
            metadata = self.metadata.to_dict()  # type: ignore
            metadata["primary_key"] = self.primary_key
            metadata["fields"] = self._get_fields_as_dict()
            json.dump(metadata, ostream, indent=2, default=str)

    @classmethod
    def read_csv(cls, path: Union[str, Path]) -> "Table":
        """
        Read the table from csv plus accompanying JSON sidecar.
        """
        if isinstance(path, Path):
            path = path.as_posix()

        if not path.endswith(".csv"):
            raise ValueError(f'filename must end in ".csv": {path}')

        # load the data
        df = Table(pd.read_csv(path, index_col=False, na_values=[""], keep_default_na=False))

        # load the metadata
        metadata = cls._read_metadata(path)

        primary_key = metadata.pop("primary_key") if "primary_key" in metadata else []
        fields = metadata.pop("fields") if "fields" in metadata else {}

        df.metadata = TableMeta(**metadata)
        df._fields = defaultdict(VariableMeta, {k: VariableMeta.from_dict(v) for k, v in fields.items()})

        if primary_key:
            df.set_index(primary_key, inplace=True)

        return df

    def update_metadata(self, **kwargs) -> "Table":
        """Set Table metadata."""
        for k, v in kwargs.items():
            assert hasattr(self.metadata, k), f"unknown metadata field {k} in TableMeta"
            setattr(self.metadata, k, v)
        return self

    @classmethod
    def _add_metadata(cls, df: pd.DataFrame, path: str) -> None:
        """Read metadata from JSON sidecar and add it to the dataframe."""
        metadata = cls._read_metadata(path)

        primary_key = metadata.get("primary_key", [])
        fields = metadata.pop("fields") if "fields" in metadata else {}

        df.metadata = TableMeta.from_dict(metadata)
        df._set_fields_from_dict(fields)

        # NOTE: setting index is really slow for large datasets
        if primary_key:
            df.set_index(primary_key, inplace=True)

    @classmethod
    def read_feather(cls, path: Union[str, Path]) -> "Table":
        """
        Read the table from feather plus accompanying JSON sidecar.

        The path may be a local file path or a URL.
        """
        if isinstance(path, Path):
            path = path.as_posix()

        if not path.endswith(".feather"):
            raise ValueError(f'filename must end in ".feather": {path}')

        # load the data and add metadata
        df = Table(pd.read_feather(path))
        cls._add_metadata(df, path)
        return df

    @classmethod
    def read_parquet(cls, path: Union[str, Path]) -> "Table":
        """
        Read the table from a parquet file plus accompanying JSON sidecar.

        The path may be a local file path or a URL.
        """
        if isinstance(path, Path):
            path = path.as_posix()

        if not path.endswith(".parquet"):
            raise ValueError(f'filename must end in ".parquet": {path}')

        # load the data and add metadata
        df = Table(pd.read_parquet(path))
        cls._add_metadata(df, path)
        return df

    def _get_fields_as_dict(self) -> Dict[str, Any]:
        return {col: self._fields[col].to_dict() for col in self.all_columns}

    def _set_fields_from_dict(self, fields: Dict[str, Any]) -> None:
        self._fields = defaultdict(VariableMeta, {k: VariableMeta.from_dict(v) for k, v in fields.items()})

    @staticmethod
    def _read_metadata(data_path: str) -> Dict[str, Any]:
        import requests

        metadata_path = splitext(data_path)[0] + ".meta.json"

        if metadata_path.startswith("http"):
            return cast(Dict[str, Any], requests.get(metadata_path).json())

        with open(metadata_path, "r") as istream:
            return cast(Dict[str, Any], json.load(istream))

    def __setitem__(self, key: Any, value: Any) -> Any:
        super().__setitem__(key, value)

        # propagate metadata when we add a series to a table
        if isinstance(key, str):
            if isinstance(value, variables.Variable):
                # variable needs to be assigned name to make VariableMeta work
                if not value.name:
                    value.name = key
                if value.name == variables.UNNAMED_VARIABLE:
                    # Update the variable name, if it had the unnamed variable tag.
                    # Replace all instances of unnamed variables in the processing log by the actual name of the new
                    # variable.
                    # WARNING: This process assumes that all instances of unnamed variable tag correspond to the new
                    #  variable.
                    variables.update_variable_name(variable=value, name=key)
                self._fields[key] = value.metadata
                # TODO: The only reason why we have to check if variables.PROCESSING_LOG is true is the test
                #   "test_field_access_can_be_typecast". Consider adapting the test and then remove this check.
                if variables.PROCESSING_LOG and len(value.metadata.processing_log) > 0:
                    # If a new variable is added to a table, check its last entry in the processing log.
                    # If the last variable name is different to the name of the new column, add an entry to the log,
                    # stating that the variable has changed name (from the old to the current one).
                    last_variable_name = value.metadata.processing_log[-1]["variable"]
                    if last_variable_name != key:
                        value.update_log(
                            parents=[last_variable_name], operation="rename", variable_name=key, inplace=True
                        )
            else:
                self._fields[key] = VariableMeta()

    def equals_table(self, table: "Table") -> bool:
        return (
            isinstance(table, Table)
            and self.metadata == table.metadata
            # By simply doing self.to_dict() == table.to_dict() may return False if the dictionaries are identical but
            # contain nans.
            and self.fillna(123456).to_dict() == table.fillna(123456).to_dict()
            and self._fields == table._fields
        )

    @rewrite_axis_style_signature(
        "mapper",
        [("copy", True), ("inplace", False), ("level", None), ("errors", "ignore")],
    )
    def rename(self, *args: Any, **kwargs: Any) -> Optional["Table"]:
        """Rename columns while keeping their metadata."""
        inplace = kwargs.get("inplace")
        old_cols = self.all_columns
        new_table = super().rename(*args, **kwargs)

        if inplace:
            new_table = self

        # construct new _fields attribute
        fields = {}
        for old_col, new_col in zip(old_cols, new_table.all_columns):
            if inplace:
                fields[new_col] = self._fields[old_col]
            else:
                fields[new_col] = self._fields[old_col].copy()

            # Update processing log.
            if old_col != new_col:
                fields[new_col].processing_log = variables.add_entry_to_processing_log(
                    processing_log=fields[new_col].processing_log,
                    variable_name=new_col,
                    parents=[old_col],
                    operation="rename",
                )

        new_table._fields = defaultdict(VariableMeta, fields)

        if inplace:
            return None
        else:
            return cast(Table, new_table)

    @property
    def all_columns(self) -> List[str]:
        "Return names of all columns in the dataset, including the index."
        combined: List[str] = filter(None, list(self.index.names) + list(self.columns))  # type: ignore
        return combined

    def get_column_or_index(self, name) -> variables.Variable:
        if name in self.columns:
            return self[name]
        elif name in self.index.names:
            return variables.Variable(self.index.get_level_values(name), name=name, metadata=self._fields[name])
        else:
            raise ValueError(f"'{name}' not found in columns or index")

    def update_metadata_from_yaml(
        self, path: Union[Path, str], table_name: str, extra_variables: Literal["raise", "ignore"] = "raise"
    ) -> None:
        """Update metadata of table and variables from a YAML file.
        :param path: Path to YAML file.
        :param table_name: Name of table, also updates this in the metadata.
        """
        from .meta import DatasetMeta
        from .utils import dynamic_yaml_load

        annot = dynamic_yaml_load(path, DatasetMeta._params_yaml(self.metadata.dataset or DatasetMeta()))

        self.metadata.short_name = table_name

        t_annot = annot["tables"][table_name]

        # validation
        if extra_variables == "raise":
            yaml_variable_names = t_annot.get("variables", {}).keys()
            table_variable_names = self.columns
            extra_variable_names = yaml_variable_names - table_variable_names
            if extra_variable_names:
                raise ValueError(f"Table {table_name} has extra variables: {extra_variable_names}")

        # update variables
        for v_short_name, v_annot in (t_annot.get("variables", {}) or {}).items():
            if v_short_name in self.columns:
                for k, v in v_annot.items():
                    # create an object out of sources
                    if k == "sources":
                        self[v_short_name].metadata.sources = [Source(**source) for source in v]
                    else:
                        setattr(self[v_short_name].metadata, k, v)

                    # we allow `- *descriptions` which needs to be flattened
                    if k == "description_key":
                        self[v_short_name].metadata.description_key = _flatten(
                            self[v_short_name].metadata.description_key
                        )

        # update table attributes
        for k, v in t_annot.items():
            if k != "variables":
                setattr(self.metadata, k, v)

    def prune_metadata(self) -> "Table":
        """Prune metadata for columns that are not in the table. This can happen after slicing
        the table by columns."""
        self._fields = defaultdict(VariableMeta, {col: self._fields[col] for col in self.all_columns})
        return self

    def copy(self, deep: bool = True) -> "Table":
        """Copy table together with all its metadata."""
        tab = super().copy(deep=deep)
        return tab.copy_metadata(self)

    def copy_metadata(self, from_table: "Table", deep: bool = False) -> "Table":
        """Copy metadata from a different table to self."""
        return copy_metadata(to_table=self, from_table=from_table, deep=deep)

    @overload
    def set_index(
        self,
        keys: Union[str, List[str]],
        *,
        inplace: Literal[True],
        **kwargs: Any,
    ) -> None:
        ...

    @overload
    def set_index(self, keys: Union[str, List[str]], *, inplace: Literal[False], **kwargs: Any) -> "Table":
        ...

    @overload
    def set_index(self, keys: Union[str, List[str]], **kwargs: Any) -> "Table":
        ...

    def set_index(
        self,
        keys: Union[str, List[str]],
        **kwargs: Any,
    ) -> Optional["Table"]:
        if isinstance(keys, str):
            keys = [keys]

        if kwargs.get("inplace"):
            super().set_index(keys, **kwargs)
            self.metadata.primary_key = keys
            return None
        else:
            t = super().set_index(keys, **kwargs)
            t.metadata.primary_key = keys
            return cast(Table, t)

    @overload
    def reset_index(self, *, inplace: Literal[True]) -> None:
        ...

    @overload
    def reset_index(self, *, inplace: Literal[False]) -> "Table":
        ...

    @overload
    def reset_index(self) -> "Table":
        ...

    def reset_index(self, *args, **kwargs) -> Optional["Table"]:  # type: ignore
        """Fix type signature of reset_index."""
        t = super().reset_index(*args, **kwargs)
        if kwargs.get("inplace"):
            return None
        else:
            # preserve metadata in _fields, calling reset_index() on a table drops it
            t._fields = self._fields
            return t  # type: ignore

    def join(self, other: Union[pd.DataFrame, "Table"], *args, **kwargs) -> "Table":
        """Fix type signature of join."""
        t = super().join(other, *args, **kwargs)

        t = t.copy_metadata(self)

        # copy variables metadata from other table
        if isinstance(other, Table):
            for k, v in other._fields.items():
                t._fields[k] = v.copy()
        return t  # type: ignore

    def _repr_html_(self):
        html = super()._repr_html_()
        return """
             <h2 style="margin-bottom: 0em"><pre>{}</pre></h2>
             <p style="font-variant: small-caps; font-size: 1.5em; font-family: sans-serif; color: grey; margin-top: -0.2em; margin-bottom: 0.2em">table</p>
             {}
        """.format(
            self.metadata.short_name, html
        )

    def merge(self, right, *args, **kwargs) -> "Table":
        return merge(left=self, right=right, *args, **kwargs)

    def melt(
        self,
        id_vars: Optional[Union[Tuple[str], List[str], str]] = None,
        value_vars: Optional[Union[Tuple[str], List[str], str]] = None,
        var_name: str = "variable",
        value_name: str = "value",
        short_name: Optional[str] = None,
        *args,
        **kwargs,
    ) -> "Table":
        return melt(
            frame=self,
            id_vars=id_vars,
            value_vars=value_vars,
            var_name=var_name,
            value_name=value_name,
            short_name=short_name,
            *args,
            **kwargs,
        )

    def pivot(
        self,
        *,
        index: Optional[Union[str, List[str]]] = None,
        columns: Optional[Union[str, List[str]]] = None,
        values: Optional[Union[str, List[str]]] = None,
        join_column_levels_with: Optional[str] = None,
        short_name: Optional[str] = None,
        **kwargs,
    ) -> "Table":
        return pivot(
            data=self,
            index=index,
            columns=columns,
            values=values,
            join_column_levels_with=join_column_levels_with,
            short_name=short_name,
            **kwargs,
        )

    def underscore(
        self,
        collision: Literal["raise", "rename", "ignore"] = "raise",
        inplace: bool = False,
        camel_to_snake: bool = False,
    ) -> "Table":
        """Convert column and index names to underscore. In extremely rare cases
        two columns might have the same underscored version. Use `collision` param
        to control whether to raise an error or append numbered suffix.

        Parameters
        ----------
        t : Table
            Table to underscore.
        collision : Literal["raise", "rename", "ignore"], optional
            How to handle collisions, by default "raise".
        inplace : bool, optional
            Whether to modify the table in place, by default False.
        camel_to_snake : bool, optional
            Whether to convert strings camelCase to snake_case, by default False.
        """
        t = self
        orig_cols = t.columns

        # underscore columns and resolve collisions
        new_cols = pd.Index([underscore(c, camel_to_snake=camel_to_snake) for c in t.columns])
        new_cols = _resolve_collisions(orig_cols, new_cols, collision)

        columns_map = {c_old: c_new for c_old, c_new in zip(orig_cols, new_cols)}
        if inplace:
            t.rename(columns=columns_map, inplace=True)
        else:
            t = t.rename(columns=columns_map)

        t.index.names = [underscore(e, camel_to_snake=camel_to_snake) for e in t.index.names]
        t.metadata.primary_key = t.primary_key
        t.metadata.short_name = underscore(t.metadata.short_name, camel_to_snake=camel_to_snake)

        # put original names as titles into metadata by default
        for c_old, c_new in columns_map.items():
            # if underscoring didn't change anything, don't add title
            if t[c_new].metadata.title is None and c_old != c_new:
                t[c_new].metadata.title = c_old

        return t

    def dropna(self, *args, **kwargs) -> "Table":
        tb = super().dropna(*args, **kwargs).copy()
        for column in list(tb.all_columns):
            tb._fields[column].processing_log = variables.add_entry_to_processing_log(
                processing_log=tb._fields[column].processing_log,
                variable_name=column,
                parents=[column],
                operation="dropna",
            )

        return cast("Table", tb)

    def update_log(
        self,
        operation: str,
        parents: Optional[List[Any]] = None,
        variable_names: Optional[List[str]] = None,
        comment: Optional[str] = None,
        inplace: bool = False,
    ) -> Optional["Table"]:
        return update_log(
            table=self,
            operation=operation,
            parents=parents,
            variable_names=variable_names,
            comment=comment,
            inplace=inplace,  # type: ignore
        )

    def amend_log(
        self,
        operation: str,
        parents: Optional[List[Any]] = None,
        variable_names: Optional[List[str]] = None,
        comment: Optional[str] = None,
        entry_num: Optional[int] = -1,
        inplace: bool = False,
    ) -> Optional["Table"]:
        return amend_log(
            table=self,
            operation=operation,
            parents=parents,
            variable_names=variable_names,
            comment=comment,
            entry_num=entry_num,
            inplace=inplace,
        )

    def sort_values(self, by: str, *args, **kwargs) -> "Table":
        tb = super().sort_values(by=by, *args, **kwargs).copy()
        for column in list(tb.all_columns):
            tb._fields[column].processing_log = variables.add_entry_to_processing_log(
                processing_log=tb._fields[column].processing_log, variable_name=column, parents=[by], operation="sort"
            )

        return cast("Table", tb)

    def sum(self, *args, **kwargs) -> variables.Variable:
        variable_name = variables.UNNAMED_VARIABLE
        variable = variables.Variable(super().sum(*args, **kwargs), name=variable_name)
        variable.metadata = variables.combine_variables_metadata(
            variables=[self[column] for column in self.columns], operation="+", name=variable_name
        )

        return variable

    def prod(self, *args, **kwargs) -> variables.Variable:
        variable_name = variables.UNNAMED_VARIABLE
        variable = variables.Variable(super().prod(*args, **kwargs), name=variable_name)
        variable.metadata = variables.combine_variables_metadata(
            variables=[self[column] for column in self.columns], operation="*", name=variable_name
        )

        return variable

    def assign(self, *args, **kwargs) -> "Table":
        return super().assign(*args, **kwargs)  # type: ignore

    def __add__(self, other: Union[Scalar, Series, variables.Variable]) -> "Table":
        tb = cast(Table, Table(super().__add__(other=other)).copy_metadata(self))
        # The following would have a parents only the scalar, not the scalar and the corresponding variable.
        # tb = update_log(table=tb, operation="+", parents=[other], variable_names=tb.columns)
        # Instead, update the processing log of each variable in the table.
        for column in tb.columns:
            tb[column] = variables.update_log(
                variable=tb[column], parents=[column, other], operation="+", variable_name=column
            )
        return tb

    def __iadd__(self, other: Union[Scalar, Series, variables.Variable]) -> "Table":
        return self.__add__(other)

    def __sub__(self, other: Union[Scalar, Series, variables.Variable]) -> "Table":
        tb = cast(Table, Table(super().__sub__(other=other)).copy_metadata(self))
        for column in tb.columns:
            tb[column] = variables.update_log(
                variable=tb[column], parents=[column, other], operation="-", variable_name=column
            )
        return tb

    def __isub__(self, other: Union[Scalar, Series, variables.Variable]) -> "Table":
        return self.__sub__(other)

    def __mul__(self, other: Union[Scalar, Series, variables.Variable]) -> "Table":
        tb = cast(Table, Table(super().__mul__(other=other)).copy_metadata(self))
        for column in tb.columns:
            tb[column] = variables.update_log(
                variable=tb[column], parents=[column, other], operation="*", variable_name=column
            )
        return tb

    def __imul__(self, other: Union[Scalar, Series, variables.Variable]) -> "Table":
        return self.__mul__(other)

    def __truediv__(self, other: Union[Scalar, Series, variables.Variable]) -> "Table":
        tb = cast(Table, Table(super().__truediv__(other=other)).copy_metadata(self))
        for column in tb.columns:
            tb[column] = variables.update_log(
                variable=tb[column], parents=[column, other], operation="/", variable_name=column
            )
        return tb

    def __itruediv__(self, other: Union[Scalar, Series, variables.Variable]) -> "Table":
        return self.__truediv__(other)

    def __floordiv__(self, other: Union[Scalar, Series, variables.Variable]) -> "Table":
        tb = cast(Table, Table(super().__floordiv__(other=other)).copy_metadata(self))
        for column in tb.columns:
            tb[column] = variables.update_log(
                variable=tb[column], parents=[column, other], operation="//", variable_name=column
            )
        return tb

    def __ifloordiv__(self, other: Union[Scalar, Series, variables.Variable]) -> "Table":
        return self.__floordiv__(other)

    def __mod__(self, other: Union[Scalar, Series, variables.Variable]) -> "Table":
        tb = cast(Table, Table(super().__mod__(other=other)).copy_metadata(self))
        for column in tb.columns:
            tb[column] = variables.update_log(
                variable=tb[column], parents=[column, other], operation="%", variable_name=column
            )
        return tb

    def __imod__(self, other: Union[Scalar, Series, variables.Variable]) -> "Table":
        return self.__mod__(other)

    def __pow__(self, other: Union[Scalar, Series, variables.Variable]) -> "Table":
        tb = cast(Table, Table(super().__pow__(other=other)).copy_metadata(self))
        for column in tb.columns:
            tb[column] = variables.update_log(
                variable=tb[column], parents=[column, other], operation="**", variable_name=column
            )
        return tb

    def __ipow__(self, other: Union[Scalar, Series, variables.Variable]) -> "Table":
        return self.__pow__(other)

    def sort_index(self, *args, **kwargs) -> "Table":
        return super().sort_index(*args, **kwargs)  # type: ignore

    def groupby(self, *args, **kwargs) -> "TableGroupBy":
        return TableGroupBy(pd.DataFrame.groupby(self.copy(deep=False), *args, **kwargs), self.metadata, self._fields)


def _create_table(df: pd.DataFrame, metadata: TableMeta, fields: Dict[str, VariableMeta]) -> Table:
    """Create a table with metadata."""
    tb = Table(df, metadata=metadata.copy())
    tb._fields = defaultdict(VariableMeta, fields)
    return tb


class TableGroupBy:
    # fixes type hints
    __annotations__ = {}

    def __init__(self, groupby: pd.core.groupby.DataFrameGroupBy, metadata: TableMeta, fields: Dict[str, Any]):
        self.groupby = groupby
        self.metadata = metadata
        self._fields = fields

    @overload
    def __getattr__(self, name: Literal["count", "size", "sum", "mean", "median"]) -> Callable[[], "Table"]:
        ...

    @overload
    def __getattr__(self, name: str) -> "VariableGroupBy":
        ...

    def __getattr__(self, name: str) -> Union[Callable[..., "Table"], "VariableGroupBy"]:
        # Calling method on the groupby object
        if isinstance(getattr(self.groupby, name), types.MethodType):

            def func(*args, **kwargs):
                """Apply function and return variable with proper metadata."""
                df = getattr(self.groupby, name)(*args, **kwargs)
                if df.ndim == 1:
                    # output is series, e.g. `size` function
                    return df
                else:
                    return _create_table(df, self.metadata, self._fields)

            self.__annotations__[name] = Callable[..., "Table"]
            return func
        else:
            self.__annotations__[name] = VariableGroupBy
            return VariableGroupBy(getattr(self.groupby, name), name, self._fields[name])

    @overload
    def __getitem__(self, key: str) -> "VariableGroupBy":
        ...

    @overload
    def __getitem__(self, key: list) -> "TableGroupBy":
        ...

    def __getitem__(self, key: Union[str, list]) -> Union["VariableGroupBy", "TableGroupBy"]:
        if isinstance(key, list):
            return TableGroupBy(self.groupby[key], self.metadata, self._fields)
        else:
            return self.__getattr__(key)  # type: ignore

    def __iter__(self) -> Iterator[Tuple[Any, "Table"]]:
        for name, group in self.groupby:
            yield name, _create_table(group, self.metadata, self._fields)

    def _groupby_operation(self, op, *args, **kwargs):
        df = getattr(self.groupby, op)(*args, **kwargs)
        return _create_table(df, self.metadata, self._fields)

    def agg(self, func: Any, *args, **kwargs) -> "Table":
        df = self.groupby.agg(func, *args, **kwargs)

        # agg returning multiindex is not yet supported
        if isinstance(df.columns, pd.MultiIndex):
            return _create_table(df, self.metadata, self._fields)
        else:
            return _create_table(df, self.metadata, self._fields)


class VariableGroupBy:
    def __init__(self, groupby: pd.core.groupby.SeriesGroupBy, name: str, metadata: VariableMeta):
        self.groupby = groupby
        self.metadata = metadata
        self.name = name

    def __getattr__(self, funcname) -> Callable[..., "Table"]:
        def func(*args, **kwargs):
            """Apply function and return variable with proper metadata."""
            # out = getattr(self.groupby, funcname)(*args, **kwargs)
            ff = getattr(self.groupby, funcname)
            out = ff(*args, **kwargs)

            # this happens when we use e.g. agg([min, max]), propagate metadata from the original then
            if isinstance(out, Table):
                out._fields = defaultdict(VariableMeta, {k: self.metadata for k in out.columns})
                return out
            elif isinstance(out, variables.Variable):
                out.metadata = self.metadata.copy()
                return out
            elif isinstance(out, pd.Series):
                return variables.Variable(out, name=self.name, metadata=self.metadata)
            else:
                raise NotImplementedError()

        return func  # type: ignore


def merge(
    left,
    right,
    how="inner",
    on=None,
    left_on=None,
    right_on=None,
    suffixes=("_x", "_y"),
    short_name: Optional[str] = None,
    **kwargs,
) -> Table:
    if ("left_index" in kwargs) or ("right_index" in kwargs):
        # TODO: Arguments left_index/right_index are not implemented.
        raise NotImplementedError(
            "Arguments 'left_index' and 'right_index' currently not implemented in function 'merge'."
        )
    # Create merged table.
    tb = Table(
        pd.merge(
            left=left, right=right, how=how, on=on, left_on=left_on, right_on=right_on, suffixes=suffixes, **kwargs
        )
    )

    # If arguments "on", "left_on", or "right_on" are given as strings, convert them to lists.
    if isinstance(on, str):
        on = [on]
    if isinstance(left_on, str):
        left_on = [left_on]
    if isinstance(right_on, str):
        right_on = [right_on]

    if (on is None) and (left_on is None):
        # By construction, either "on" is passed, or both "left_on" and "right_on".
        # Any other possibility will raise a MergeError, and hence doesn't need to be considered.
        # If none of them is not specified, assume we are joining on common columns.
        on = list(set(left.all_columns) & set(right.all_columns))
        left_on = []
        right_on = []
    elif left_on is not None:
        # By construction, right_on must also be given.
        on = []
    else:
        # Here, "on" is given, but not "left_on".
        left_on = []
        right_on = []

    # Find columns that existed in both left and right tables whose name will be modified with suffixes.
    overlapping_columns = ((set(left.all_columns) - set(left_on)) & (set(right.all_columns) - set(right_on))) - set(on)  # type: ignore

    # Find columns that existed in both left and right tables that will preserve their names (since they are columns to join on).
    common_columns = on or (set(left_on) & set(right_on))  # type: ignore

    columns_from_left = set(left.all_columns) - set(common_columns)
    columns_from_right = set(right.all_columns) - set(common_columns)

    for column in columns_from_left:
        if column in overlapping_columns:
            new_column = f"{column}{suffixes[0]}"
        else:
            new_column = column
        tb[new_column].metadata = variables.combine_variables_metadata([left[column]], operation="merge", name=column)

    for column in columns_from_right:
        if column in overlapping_columns:
            new_column = f"{column}{suffixes[1]}"
        else:
            new_column = column
        tb[new_column].metadata = variables.combine_variables_metadata([right[column]], operation="merge", name=column)

    for column in common_columns:
        tb[column].metadata = variables.combine_variables_metadata(
            [left[column], right[column]], operation="merge", name=column
        )

    # Update table metadata.
    tb.metadata = combine_tables_metadata(tables=[left, right], short_name=short_name)

    return tb


def concat(
    objs: List[Table],
    *,
    axis: Union[int, str] = 0,
    join: str = "outer",
    ignore_index: bool = False,
    short_name: Optional[str] = None,
    **kwargs,
) -> Table:
    # TODO: Add more logic to this function to handle indexes and possibly other arguments.
    table = Table(pd.concat(objs=objs, axis=axis, join=join, ignore_index=ignore_index, **kwargs))  # type: ignore

    if (axis == 1) or (axis == "columns"):
        # Original function pd.concat allows returning a dataframe with multiple columns with the same name.
        # But this should not be allowed (and metadata cannot be stored for different variables if they have the same name).
        repeated_columns = table.columns[table.columns.duplicated()].tolist()
        if len(repeated_columns) > 0:
            raise KeyError(f"Concatenated table contains repeated columns: {repeated_columns}")

        # Assign variable metadata from input tables.
        for table_i in objs:
            for column in table_i.columns:
                table[column].metadata = table_i[column].metadata

    else:
        if list(filter(None, table.index.names)):
            raise NotImplementedError("Concatenation of tables with index is not implemented.")

        # Add to each column either the metadata of the original variable (if the variable appeared only in one of the input
        # tables) or the combination of the metadata from different tables (if the variable appeared in various tables).
        for column in table.all_columns:
            variables_to_combine = [table_i[column] for table_i in objs if column in table_i.all_columns]
            table._fields[column] = variables.combine_variables_metadata(
                variables=variables_to_combine, operation="concat", name=column
            )

    # Update table metadata.
    table.metadata = combine_tables_metadata(tables=objs, short_name=short_name)

    return table


def melt(
    frame: Table,
    id_vars: Optional[Union[Tuple[str], List[str], str]] = None,
    value_vars: Optional[Union[Tuple[str], List[str], str]] = None,
    var_name: str = "variable",
    value_name: str = "value",
    short_name: Optional[str] = None,
    *args,
    **kwargs,
) -> Table:
    # TODO: We may need to implement some mor logic here to handle multi-index dataframes.
    # Get the new melt table.
    table = Table(
        pd.melt(
            frame=frame,
            id_vars=id_vars,
            value_vars=value_vars,
            var_name=var_name,
            value_name=value_name,
            *args,
            **kwargs,
        )
    )

    # Get the list of column names used as id variables.
    if id_vars is None:
        id_vars_list = []
    elif isinstance(id_vars, str):
        id_vars_list: List[str] = [id_vars]
    else:
        id_vars_list = id_vars  # type: ignore

    # Get the list of column names used as id and value variables.
    if value_vars is None:
        value_vars_list = [column for column in frame.columns if column not in id_vars_list]
    elif isinstance(value_vars, str):
        value_vars_list = [value_vars]
    else:
        value_vars_list = value_vars  # type: ignore

    # Combine metadata of value variables and assign the combination to the new "value" column.
    table[value_name].metadata = variables.combine_variables_metadata(
        variables=[frame[var] for var in value_vars_list], operation="melt", name=value_name
    )

    # Assign that combined metadata also to the new "variable" column.
    table[var_name].metadata = variables.combine_variables_metadata(
        variables=[frame[var] for var in value_vars_list], operation="melt", name=var_name
    )

    for variable in id_vars_list:
        # Combine metadata of id variables and assign the combination to the new "id" variable.
        table[variable].metadata = variables.combine_variables_metadata(
            variables=[frame[variable]], operation="melt", name=variable
        )

    # Update table metadata.
    table.metadata = combine_tables_metadata(tables=[frame], short_name=short_name)

    return table


def _flatten_multiindex_column_names(table: Table, join_column_levels_with: str) -> List[str]:
    new_columns = []
    for column in table.columns:
        if isinstance(column, tuple):
            levels = [str(level) for level in column if len(str(level)) > 0]
            new_column = join_column_levels_with.join(levels)
        else:
            new_column = column
        new_columns.append(new_column)

    return new_columns


def pivot(
    data: Table,
    *,
    index: Optional[Union[str, List[str]]] = None,
    columns: Optional[Union[str, List[str]]] = None,
    values: Optional[Union[str, List[str]]] = None,
    join_column_levels_with: Optional[str] = None,
    short_name: Optional[str] = None,
    **kwargs,
) -> Table:
    # Get the new pivot table.
    table = Table(
        pd.pivot(
            data=data,
            index=index,
            columns=columns,
            values=values,
            **kwargs,
        )
    )

    # Update variable metadata in the new table.
    for column in table.columns:
        if isinstance(values, str):
            column_name = values
        else:
            column_name = column[0]
        variables_to_combine = [data[column_name]]
        # "column" is a tuple with all column index levels.
        # For now, I assume the only metadata we want to propagate is the one of the upper level.
        # Alternatively, we could combine the metadata of the upper level variable with the metadata of the original
        # variable of all subsequent levels.
        column_metadata = variables.combine_variables_metadata(
            variables=variables_to_combine, operation="pivot", name=column_name
        )
        # Assign metadata of the original variable in the upper level to the new multiindex column.
        # NOTE: This allows accessing the metadata via, e.g. `table[("level_0", "level_1", "level_2")].metadata`,
        # but not via, e.g. `table["level_0"]["level_1"]["level_2"].metadata`.
        # There may be a way to allow for both.
        table[column].metadata = column_metadata

    # Transfer also the metadata of the index columns.
    # Note: This metadata will only be accessible if columns are reset and flattened to one level.
    for index_column in list(table.index.names):
        table._fields[index_column] = data._fields[index_column]

    if join_column_levels_with is not None:
        # Gather metadata of index columns.
        index_metadata = [table._fields[index_column] for index_column in table.index.names]
        # Gather metadata of each multiindex column.
        columns_metadata = [table[column].metadata for column in table.columns]
        # Reset index (which can create multi-index columns).
        table = table.reset_index()
        # Join column levels with a certain string, e.g. ("level_0", "level_1", "level_2") -> "level_0-level_1-level_2".
        table.columns = _flatten_multiindex_column_names(table, join_column_levels_with=join_column_levels_with)
        # Assign the gathered metadata for indexes and columns to the corresponding new columns.
        for i, column in enumerate(table.columns):
            table[column].metadata = (index_metadata + columns_metadata)[i]

    # Update table metadata.
    table.metadata = combine_tables_metadata(tables=[data], short_name=short_name)

    return table


def _add_table_and_variables_metadata_to_table(table: Table, metadata: Optional[TableMeta]) -> Table:
    if metadata is not None:
        table.metadata = metadata
        for column in list(table.all_columns):
            if "origins" in metadata.dataset.to_dict():  # type: ignore
                table._fields[column].origins = metadata.dataset.origins  # type: ignore
            else:
                table._fields[column].sources = metadata.dataset.sources  # type: ignore
            table._fields[column].licenses = metadata.dataset.licenses  # type: ignore
    table = update_processing_logs_when_loading_or_creating_table(table=table)

    return table


def read_csv(
    filepath_or_buffer: Union[str, Path, IO[AnyStr]],
    metadata: Optional[TableMeta] = None,
    underscore: bool = False,
    *args,
    **kwargs,
) -> Table:
    table = Table(pd.read_csv(filepath_or_buffer=filepath_or_buffer, *args, **kwargs), underscore=underscore)
    table = _add_table_and_variables_metadata_to_table(table=table, metadata=metadata)
    if isinstance(filepath_or_buffer, (str, Path)):
        table = update_log(table=table, operation="load", parents=[filepath_or_buffer])
    else:
        log.warning("Currently, the processing log cannot be updated unless you pass a path to read_csv.")

    return cast(Table, table)


def read_feather(
    filepath: Union[str, Path],
    metadata: Optional[TableMeta] = None,
    underscore: bool = False,
    *args,
    **kwargs,
) -> Table:
    table = Table(pd.read_feather(filepath, *args, **kwargs), underscore=underscore)
    table = _add_table_and_variables_metadata_to_table(table=table, metadata=metadata)
    table = update_log(table=table, operation="load", parents=[filepath])

    return cast(Table, table)


def read_excel(
    io: Union[str, Path], *args, metadata: Optional[TableMeta] = None, underscore: bool = False, **kwargs
) -> Table:
    assert not isinstance(kwargs.get("sheet_name"), list), "Argument 'sheet_name' must be a string or an integer."
    table = Table(pd.read_excel(io=io, *args, **kwargs), underscore=underscore)
    table = _add_table_and_variables_metadata_to_table(table=table, metadata=metadata)
    # Note: Maybe we should include the sheet name in parents.
    table = update_log(table=table, operation="load", parents=[io], inplace=False)

    return cast(Table, table)


def read_from_records(data: Any, *args, metadata: Optional[TableMeta] = None, underscore: bool = False, **kwargs):
    table = Table(pd.DataFrame.from_records(data=data, *args, **kwargs), underscore=underscore)
    table = _add_table_and_variables_metadata_to_table(table=table, metadata=metadata)
    # NOTE: Parents could be passed as arguments, or extracted from metadata.
    table = update_log(table=table, operation="load", parents=["local_data"], inplace=False)

    return table


def read_from_dict(
    data: Dict[Any, Any], *args, metadata: Optional[TableMeta] = None, underscore: bool = False, **kwargs
) -> Table:
    table = Table(pd.DataFrame.from_dict(data=data, *args, **kwargs), underscore=underscore)
    table = _add_table_and_variables_metadata_to_table(table=table, metadata=metadata)
    # NOTE: Parents could be passed as arguments, or extracted from metadata.
    table = update_log(table=table, operation="load", parents=["local_data"], inplace=False)

    return table


def read_json(
    path_or_buf: Union[str, Path, IO[AnyStr]],
    metadata: Optional[TableMeta] = None,
    underscore: bool = False,
    *args,
    **kwargs,
) -> Table:
    table = Table(pd.read_json(path_or_buf=path_or_buf, *args, **kwargs), underscore=underscore)
    table = _add_table_and_variables_metadata_to_table(table=table, metadata=metadata)
    if isinstance(path_or_buf, (str, Path)):
        table = update_log(table=table, operation="load", parents=[path_or_buf])
    else:
        log.warning("Currently, the processing log cannot be updated unless you pass a path to read_json.")

    return cast(Table, table)


class ExcelFile(pd.ExcelFile):
    def parse(
        self,
        sheet_name: Union[str, int] = 0,
        *args,
        metadata: Optional[TableMeta] = None,
        underscore: bool = False,
        **kwargs,
    ):
        # Note: Maybe we should include the sheet name in parents.
        df = super().parse(sheet_name=sheet_name, *args, **kwargs)  # type: ignore
        table = Table(df, underscore=underscore, short_name=str(sheet_name))
        if metadata is not None:
            table = _add_table_and_variables_metadata_to_table(table=table, metadata=metadata)
        # Note: Maybe we should include the sheet name in parents.
        table = update_log(table=table, operation="load", parents=[self.io], inplace=False)

        return table


def update_processing_logs_when_loading_or_creating_table(table: Table) -> Table:
    # Add entry to processing log, specifying that each variable was loaded from this table.
    try:
        # If the table comes from an ETL dataset, generate a URI for the table.
        table_uri = f"{table.metadata.dataset.uri}/{table.metadata.short_name}"  # type: ignore
        parents = [table_uri]
        operation = "load"
    except (AssertionError, AttributeError):
        # The table doesn't have an uri, which means it was probably created from scratch.
        parents = []
        operation = "create"

    table = _add_processing_log_entry_to_each_variable(table=table, parents=parents, operation=operation)

    return table


def update_processing_logs_when_saving_table(table: Table, path: Union[str, Path]) -> Table:
    # Infer the ETL uri from the path where the table will be saved.
    # Note: If the path does not fit the expected format, the result will be an arbitrary path, but it will not raise an
    # error, as long as path is a Path.
    path = Path(path)
    uri = "/".join(path.absolute().parts[-5:-1] + tuple([path.stem]))
    table = _add_processing_log_entry_to_each_variable(table=table, parents=[uri], operation="save")

    return table


def _add_processing_log_entry_to_each_variable(
    table: Table, parents: List[Any], operation: variables.OPERATION
) -> Table:
    # Add a processing log entry to each column, including index columns.
    for column in list(table.all_columns):
        # New entry to add to the processing log.
        # Note: The function add_entry_to_processing_log receives the argument variable_name, but in the processing log,
        # the entry has the field "variable" (for simplicity).
        log_new_entry = {"variable_name": column, "parents": parents, "operation": operation}

        if log_new_entry not in table._fields[column].processing_log:
            # If the processing log is not empty but the last entry is identical to the one we want to insert, skip, to
            # avoid storing the same entry multiple times.
            # This happens for example when saving tables, given that tables are stored in different formats.
            # Otherwise, append a new entry to the processing log.
            table._fields[column].processing_log = variables.add_entry_to_processing_log(
                processing_log=table._fields[column].processing_log, **log_new_entry
            )

    return table


def assign_dataset_sources_origins_and_licenses_to_each_variable(table: Table) -> Table:
    # Get sources, origins and licenses from the table dataset.
    sources = []
    origins = []
    licenses = []
    if hasattr(table.metadata, "dataset") and hasattr(table.metadata.dataset, "sources"):
        sources = table.metadata.dataset.sources  # type: ignore
    if hasattr(table.metadata, "dataset") and hasattr(table.metadata.dataset, "origins"):
        origins = table.metadata.dataset.origins  # type: ignore
    if hasattr(table.metadata, "dataset") and hasattr(table.metadata.dataset, "sources"):
        licenses = table.metadata.dataset.licenses  # type: ignore

    if len(sources) == len(licenses) == len(origins) == 0:
        # There are no default sources/origins/licenses to assign to each variable.
        return table

    # If a variable does not have sources/origins/licenses defined, assign the ones from the dataset.
    # Do this for all columns, including index columns.
    for column in list(table.all_columns):
        if len(table._fields[column].sources) == 0:
            table._fields[column].sources = sources
        if len(table._fields[column].origins) == 0:
            table._fields[column].origins = origins
        if len(table._fields[column].licenses) == 0:
            table._fields[column].licenses = licenses

    return table


def copy_metadata(from_table: Table, to_table: Table, deep=False) -> Table:
    """Copy metadata from a different table to self."""
    tab = Table(pd.DataFrame.copy(to_table, deep=deep), metadata=from_table.metadata.copy())

    common_columns = set(to_table.all_columns) & set(from_table.all_columns)

    new_fields = defaultdict(VariableMeta)
    for k in common_columns:
        # copy if we have metadata in the other table
        if k in from_table._fields:
            new_fields[k] = from_table._fields[k].copy()
        # otherwise keep current metadata (if it exists)
        elif k in to_table._fields:
            new_fields[k] = to_table._fields[k]

    tab._fields = new_fields
    return tab


@overload
def update_log(
    table: Table,
    operation: str,
    parents: Optional[List[Any]] = None,
    variable_names: Optional[List[str]] = None,
    comment: Optional[str] = None,
    inplace: Literal[True] = True,
) -> None:
    ...


@overload
def update_log(
    table: Table,
    operation: str,
    parents: Optional[List[Any]] = None,
    variable_names: Optional[List[str]] = None,
    comment: Optional[str] = None,
    inplace: Literal[False] = False,
) -> Table:
    ...


def update_log(
    table: Table,
    operation: str,
    parents: Optional[List[Any]] = None,
    variable_names: Optional[List[str]] = None,
    comment: Optional[str] = None,
    inplace: bool = False,
) -> Optional[Table]:
    if not inplace:
        table = table.copy()

    # Append a new entry to the processing log of the required variables.
    if variable_names is None:
        # If no variable is specified, assume all (including index columns).
        variable_names = list(table.all_columns)
    for column in variable_names:
        # If parents is not defined, assume the parents are simply the current variable.
        _parents = parents or [column]
        # Update (in place) the processing log of current variable.
        table._fields[column].processing_log = variables.add_entry_to_processing_log(
            processing_log=table._fields[column].processing_log,
            variable_name=column,
            parents=_parents,
            operation=operation,
            comment=comment,
        )

    if not inplace:
        return table


def amend_log(
    table: Table,
    operation: str,
    parents: Optional[List[Any]] = None,
    variable_names: Optional[List[str]] = None,
    comment: Optional[str] = None,
    entry_num: Optional[int] = -1,
    inplace: bool = False,
) -> Optional[Table]:
    if not inplace:
        table = table.copy()

    # Append a new entry to the processing log of the required variables.
    if variable_names is None:
        # If no variable is specified, assume all (including index columns).
        variable_names = list(table.all_columns)
    for column in variable_names:
        # If parents is not defined, assume the parents are simply the current variable.
        _parents = parents or [column]
        # Update (in place) the processing log of current variable.
        table._fields[column].processing_log = variables.amend_entry_in_processing_log(
            processing_log=table._fields[column].processing_log,
            variable_name=column,
            parents=_parents,
            operation=operation,
            comment=comment,
            entry_num=entry_num,
        )

    if not inplace:
        return table


def get_unique_sources_from_tables(tables: List[Table]) -> List[Source]:
    # Make a list of all sources of all variables in all tables.
    sources = sum([table._fields[column].sources for table in tables for column in list(table.all_columns)], [])

    # Get unique array of tuples of source fields (respecting the order).
    return pd.unique(sources).tolist()


def get_unique_origins_from_tables(tables: List[Table]) -> List[Origin]:
    # First, ensure only "origins" (and not "sources") are used in the variables.
    if any(
        [
            ("sources" in table[column].metadata.to_dict()) and (len(table[column].metadata.sources) > 0)
            for table in tables
            for column in table.columns
        ]
    ):
        log.warning("Origins and sources mixed in tables. The resulting dataset may not include all origins.")

    # Make a list of all origins of all variables in all tables.
    origins = sum([table._fields[column].origins for table in tables for column in list(table.all_columns)], [])

    # Get unique array of tuples of source fields (respecting the order).
    return pd.unique(origins).tolist()


def get_unique_licenses_from_tables(tables: List[Table]) -> List[License]:
    # Make a list of all licenses of all variables in all tables.
    licenses = sum([table._fields[column].licenses for table in tables for column in list(table.all_columns)], [])

    # Get unique array of tuples of source fields (respecting the order).
    return pd.unique(licenses).tolist()


def _get_metadata_value_from_tables_if_all_identical(tables: List[Table], field: str) -> Optional[Any]:
    # Get unique values from list, ignoring Nones.
    unique_values = set(
        [getattr(table.metadata, field) for table in tables if getattr(table.metadata, field) is not None]
    )
    if len(unique_values) == 1:
        combined_value = unique_values.pop()
    else:
        combined_value = None

    return combined_value


def combine_tables_title(tables: List[Table]) -> Optional[str]:
    return _get_metadata_value_from_tables_if_all_identical(tables=tables, field="title")


def combine_tables_description(tables: List[Table]) -> Optional[str]:
    return _get_metadata_value_from_tables_if_all_identical(tables=tables, field="description")


def combine_tables_metadata(tables: List[Table], short_name: Optional[str] = None) -> TableMeta:
    title = combine_tables_title(tables=tables)
    description = combine_tables_description(tables=tables)
    if short_name is None:
        # If a short name is not specified, take it from the first table.
        short_name = tables[0].metadata.short_name
    metadata = TableMeta(title=title, description=description, short_name=short_name)

    return metadata


def check_all_variables_have_metadata(tables: List[Table], fields: Optional[List[str]] = None) -> None:
    if fields is None:
        fields = ["origins"]

    for table in tables:
        table_name = table.metadata.short_name
        for column in table.columns:
            for field in fields:
                if not getattr(table[column].metadata, field):
                    log.warning(f"Table {table_name}, column {column} has no {field}.")


def _flatten(lst: List[Any]) -> List[str]:
    """Flatten list that contains either strings or lists."""
    return [item for sublist in lst for item in ([sublist] if isinstance(sublist, str) else sublist)]


def _resolve_collisions(
    orig_cols: pd.Index,
    new_cols: pd.Index,
    collision: Literal["raise", "rename", "ignore"],
) -> pd.Index:
    new_cols = new_cols.copy()
    vc = new_cols.value_counts()

    colliding_cols = list(vc[vc >= 2].index)
    for colliding_col in colliding_cols:
        ixs = np.where(new_cols == colliding_col)[0]
        if collision == "raise":
            raise NameError(
                f"Columns `{orig_cols[ixs[0]]}` and `{orig_cols[ixs[1]]}` are given the same name "
                f"`{colliding_cols[0]}` after underscoring`"
            )
        elif collision == "rename":
            # give each column numbered suffix
            for i, ix in enumerate(ixs):
                new_cols.values[ix] = f"{new_cols[ix]}_{i + 1}"
        elif collision == "ignore":
            pass
        else:
            raise NotImplementedError()
    return new_cols



#
#  meta.py
#
#  Metadata helpers.
#

import dataclasses
import datetime as dt
import json
import re
from dataclasses import dataclass, field, is_dataclass
from pathlib import Path
from typing import Any, Dict, List, Literal, NewType, Optional, Union

import pandas as pd
from dataclasses_json import dataclass_json

from .utils import pruned_json

SOURCE_EXISTS_OPTIONS = Literal["fail", "append", "replace"]


YearDateLatest = NewType("YearDateLatest", str)


@pruned_json
@dataclass_json
@dataclass
class License:
    name: Optional[str] = None
    url: Optional[str] = None

    def __hash__(self):
        """Hash that uniquely identifies a License."""
        return _hash_dataclass(self)

    def to_dict(self) -> Dict[str, Any]:
        ...

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "License":
        ...

    def __bool__(self):
        return bool(self.name or self.url)


# DEPRECATED: use Origin instead
@pruned_json
@dataclass_json
@dataclass
class Source:
    """Notes on importing sources to grapher:
    - Field `source.description` gets mapped to `Internal notes`, but we rather use it for `additional_info`
    - The most important fields are `published_by` and `additional_info`
    - In admin for dataset (i.e. /admin/datasets/1234) only the first source of a dataset is shown and
        can be edited. The other ones are not visible.
    """

    name: Optional[str] = None
    description: Optional[str] = None
    url: Optional[str] = None
    source_data_url: Optional[str] = None
    owid_data_url: Optional[str] = None
    date_accessed: Optional[str] = None
    publication_date: Optional[str] = None
    publication_year: Optional[int] = None
    # specific fields for grapher
    # NOTE: it's not clear how to map description & name to fields in grapher, so
    # we're keeping both for the time being. We might consolidate them in the future
    published_by: Optional[str] = None

    def __hash__(self):
        """Hash that uniquely identifies a source."""
        return _hash_dataclass(self)

    def to_dict(self) -> Dict[str, Any]:
        ...

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "Source":
        ...

    def update(self, **kwargs: Dict[str, Any]) -> None:
        for key, value in kwargs.items():
            if value is not None:
                setattr(self, key, value)


@pruned_json
@dataclass_json
@dataclass
class Origin:
    # Producer name
    # Name of the institution or the author(s) that produced the data product.
    producer: str
    # Title of the original data product
    title: str
    # Description of the data product
    description: Optional[str] = None
    # Title of the snapshot
    # Subset of data that we extract from the data product. Only fill if it does not coincide with the title of the data product.
    title_snapshot: Optional[str] = None
    # Description of the snapshot
    # Subset of data that we extract from the data product). Only when the data product and the snapshot do not coincide, the description_snapshot
    # will contain additional information to the description of the data product.
    description_snapshot: Optional[str] = None
    # The full citation
    citation_full: Optional[str] = None
    # These will be often empty and then producer is used instead, but for the (relatively common) cases
    # where the data product is more famous than the authors we would use this (e.g. VDEM instead of the first authors)
    attribution: Optional[str] = None
    attribution_short: Optional[str] = None
    # This is also often empty but if not then it will be part of the short citation (e.g. for VDEM)
    version_producer: Optional[str] = None
    # The authorative URL of the dataset
    url_main: Optional[str] = None
    # Direct URL to download the dataset
    url_download: Optional[str] = None
    # Date when the dataset was accessed
    date_accessed: Optional[str] = None
    # Publication date or, if the exact date is not known, publication year
    date_published: Optional[YearDateLatest] = None
    # License of the dataset
    license: Optional[License] = None

    def __hash__(self):
        """Hash that uniquely identifies an origin."""
        return _hash_dataclass(self)

    def __post_init__(self):
        if self.date_published:
            # convert date or int to string
            if isinstance(self.date_published, (dt.date, int)):
                self.date_published = YearDateLatest(str(self.date_published))

            if self.date_published != "latest" and not is_year_or_date(self.date_published):
                raise ValueError("date_published should be either a year or a date or latest")

    def to_dict(self) -> Dict[str, Any]:
        ...

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "Origin":
        ...

    def update(self, **kwargs: Dict[str, Any]) -> None:
        for key, value in kwargs.items():
            if value is not None:
                setattr(self, key, value)


# Minor is for cases where we only harmonized the countries or similar
# Major is for cases where we do more, like create new aggregations, combine multiple indicators, etc.
PROCESSING_LEVELS = Literal["minor", "major"]

# Hierarchy of processing levels.
PROCESSING_LEVELS_ORDER = {
    "minor": 1,
    "major": 2,
}


@pruned_json
@dataclass_json
@dataclass
class FaqLink:
    gdoc_id: str
    fragment_id: str


GrapherConfig = Dict[str, Any]


@pruned_json
@dataclass_json
@dataclass
class VariablePresentationMeta:
    # Any fields of grapher config can be set here - title and subtitle *should* be set whenever possible
    grapher_config: Optional[GrapherConfig] = None
    # The text for the header of the data page
    title_public: Optional[str] = None
    # Shown next to title to differentiate similar indicators e.g. "future projections" or "historical values"
    title_variant: Optional[str] = None
    # Shown next to title to differentiate similar indicators e.g. "WHO" or "IHME"
    attribution_short: Optional[str] = None
    # A short text to use to credit the source e.g. at the bottom of charts. Autofilled from the list of origins (see below). Semicolon separated if there are multiple.
    attribution: Optional[str] = None
    # List of topic tags
    topic_tags: List[str] = field(default_factory=list)

    # Fields that are more work to add but of high value

    # List of google doc ids + fragment id
    faqs: List[FaqLink] = field(default_factory=list)

    def __hash__(self):
        """Hash that uniquely identifies VariablePresentationMeta."""
        return _hash_dataclass(self)

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "VariablePresentationMeta":
        ...


@pruned_json
@dataclass_json
@dataclass
class VariableMeta:
    """Allowed fields for `display` attribute used for grapher:
        name
        zeroDay
        yearIsDay
        includeInTable
        numDecimalPlaces
        conversionFactor
        entityAnnotationsMap
    Fields `unit` and `shortUnit` are copied from attributes `unit` and `short_unit`
    on VariableMeta object

    NOTE: consider using its own object for `display` instead of dict and also possibly
    underscoring fields and converting them back to camelCase before inserting to grapher
    """

    title: Optional[str] = None
    # This shouldn't be used for data pages, use `description_short`, `description_key` or `description_processing` instead
    description: Optional[str] = None
    # A 1-2 sentence description - used internally or as fallback for description_key
    description_short: Optional[str] = None
    # How did the origin describe this variable?
    description_from_producer: Optional[str] = None
    # List of bullet points for the description key (can use markdown formatting)
    description_key: List[str] = field(default_factory=list)
    origins: List[Origin] = field(default_factory=list)  # Origins is the new replacement for sources
    # Use of `licenses` is discouraged, they should be captured in origins.
    licenses: List[License] = field(default_factory=list)
    unit: Optional[str] = None
    short_unit: Optional[str] = None
    # We keep display for the time being as the "less powerful sibling" of grapherConfig below
    display: Optional[Dict[str, Any]] = None
    additional_info: Optional[Dict[str, Any]] = None  # Only used for internal bookkeeping

    # How much processing did we do to this data?
    processing_level: Optional[PROCESSING_LEVELS] = None
    # List of processing steps, in the future autogenerated
    processing_log: List[Dict[str, Any]] = field(default_factory=list)

    presentation: Optional[VariablePresentationMeta] = None

    # A short summary of what was done to process this indicator
    description_processing: Optional[str] = None

    # This one is the license that we give the data. Normally it will be empty and then it will
    # be our usual license (CC-BY) but in cases where special restriction apply this is where
    # we would capture this.
    license: Optional[License] = None

    # This is the old sources that we keep for compatibility. Use is strongly discouraged going forward
    sources: List[Source] = field(default_factory=list)

    def __hash__(self):
        """Hash that uniquely identifies VariableMeta."""
        return _hash_dataclass(self)

    def to_dict(self) -> Dict[str, Any]:
        ...

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "VariableMeta":
        ...

    @property
    def schema_version(self) -> int:
        """Schema version is used to easily understand everywhere what metadata standard was used
        for authoring this variable metadata. Defaults to 1 for our legacy variables. "Modern" variables
        that fill in the presentation key and use origins should record 2 here.
        """
        if self.origins or self.presentation:
            return 2
        else:
            return 1

    def _repr_html_(self):
        # Render a nice display of the table metadata
        record = self.to_dict()
        return """
             <h2 style="margin-bottom: 0em"><pre>{}</pre></h2>
             <p style="font-variant: small-caps; font-family: sans-serif; font-size: 1.5em; color: grey; margin-top: -0.2em; margin-bottom: 0.2em">variable meta</p>
             {}
        """.format(
            getattr(self, "_name", None), to_html(record)
        )

    def copy(self, deep=True) -> "VariableMeta":
        """Return a copy of the VariableMeta object."""
        if not deep:
            return dataclasses.replace(self)
        else:
            return _deepcopy_dataclass(self)


@pruned_json
@dataclass_json
@dataclass
class DatasetMeta:
    """
    The metadata for this entire dataset kept in JSON (e.g. mydataset/index.json).

    The number of fields is limited, but should handle everything that we get from
    Walden. There is a lot more opportunity to store more metadata at the table and
    the variable level.
    """

    channel: Optional[str] = None
    namespace: Optional[str] = None
    # NOTE: short_name should be underscore and validate in setter, however this
    # is nontrivial to do with `dataclass_json` (see https://github.com/lidatong/dataclasses-json/issues/176)
    short_name: Optional[str] = None
    title: Optional[str] = None
    description: Optional[str] = None
    origins: List[Origin] = field(default_factory=list)
    # sources is deprecated, use origins instead
    sources: List[Source] = field(default_factory=list)
    licenses: List[License] = field(default_factory=list)
    is_public: bool = True
    additional_info: Optional[Dict[str, Any]] = None
    version: Optional[str] = None
    # update period in days
    update_period_days: Optional[str] = None

    # an md5 checksum of the ingredients used to make this dataset
    source_checksum: Optional[str] = None

    def __hash__(self):
        """Hash that uniquely identifies DatasetMeta."""
        return _hash_dataclass(self)

    def __post_init__(self) -> None:
        """Imply version from publication_date or publication_year if not given
        in __init__."""
        if self.version is None:
            if len(self.sources) == 1:
                (source,) = self.sources
                if source.publication_date:
                    self.version = str(source.publication_date)
                elif source.publication_year:
                    self.version = str(source.publication_year)
                else:
                    self.version = None

    def save(self, filename: Union[str, Path]) -> None:
        filename = Path(filename).as_posix()
        with open(filename, "w") as ostream:
            json.dump(self.to_dict(), ostream, indent=2, default=str)

    @classmethod
    def load(cls, filename: str) -> "DatasetMeta":
        with open(filename) as istream:
            return cls.from_dict(json.load(istream))

    def to_dict(self) -> Dict[str, Any]:
        ...

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "DatasetMeta":
        ...

    def _params_yaml(self) -> dict:
        """Parameters passed to YAML for dynamic interpolation."""
        params = {}
        if self.version and self.version != "latest":
            params["YEAR"] = pd.to_datetime(self.version).year
        return params

    def update_from_yaml(self, path: Union[Path, str], if_source_exists: SOURCE_EXISTS_OPTIONS = "fail") -> None:
        """The main reason for wanting to do this is to manually override what goes into Grapher before an export."""
        from owid.catalog import utils

        annot = utils.dynamic_yaml_load(path, self._params_yaml())

        dataset_sources = annot.get("dataset", {}).get("sources", []) or []

        # update sources of dataset, if there are no sources in the new dataset, don't update existing ones
        if if_source_exists == "replace" and dataset_sources:
            self.sources = []

        new_sources = []
        for source_annot in dataset_sources:
            # if there's an existing source, update it
            ds_sources = [s for s in self.sources if s.name == source_annot["name"]]
            if ds_sources:
                ds_sources[0].update(**source_annot)
            # there is already a source in a dataset, raise an error
            elif self.sources and if_source_exists == "fail":
                raise ValueError(f"Source {self.sources[0].name} would be overwritten by source {source_annot['name']}")
            # otherwise append it
            else:
                new_sources.append(Source(**source_annot))

        self.sources.extend(new_sources)

        # update dataset
        for k, v in annot.get("dataset", {}).items():
            if k != "sources":
                setattr(self, k, v)

    @property
    def uri(self) -> str:
        """Return unique URI for this dataset if"""
        assert self.channel, "DatasetMeta.channel is not set"
        assert self.namespace, "DatasetMeta.namespace is not set"
        assert self.version, "DatasetMeta.version is not set"
        assert self.short_name, "DatasetMeta.short_name is not set"
        return f"{self.channel}/{self.namespace}/{self.version}/{self.short_name}"


@pruned_json
@dataclass_json
@dataclass
class TableMeta:
    # data about this table
    short_name: Optional[str] = None
    title: Optional[str] = None
    description: Optional[str] = None

    # a reference back to the dataset
    dataset: Optional[DatasetMeta] = field(compare=False, default=None)
    primary_key: List[str] = field(default_factory=list)

    def __hash__(self):
        """Hash that uniquely identifies TableMeta."""
        return _hash_dataclass(self)

    @property
    def checked_name(self) -> str:
        if not self.short_name:
            raise Exception("table has no short_name")

        return self.short_name

    def to_dict(self) -> Dict[str, Any]:
        ...

    @staticmethod
    def from_dict(dict: Dict[str, Any]) -> "TableMeta":
        ...

    def _repr_html_(self):
        # Render a nice display of the table metadata
        record = self.to_dict()
        short_name = record.pop("short_name")
        return """
             <h2 style="margin-bottom: 0em"><pre>{}</pre></h2>
             <p style="font-variant: small-caps; font-family: sans-serif; font-size: 1.5em; color: grey; margin-top: -0.2em; margin-bottom: 0.2em">table meta</p>
             {}
        """.format(
            short_name, to_html(record)
        )

    def copy(self, deep=True) -> "TableMeta":
        """Return a copy of the TableMeta object."""
        if not deep:
            return dataclasses.replace(self)
        else:
            return _deepcopy_dataclass(self)


def to_html(record: Any) -> Optional[str]:
    if isinstance(record, dict):
        rows = []
        for k, v in record.items():
            if not v:
                continue
            v_str = to_html(v)
            rows.append(
                """<tr><th style="text-align: right; font-family: sans-serif; vertical-align: top; padding: 0.2em 1em;"><strong>{}</strong></th><td style="text-align: left; padding: 0.2em 1em;">{}</td></tr>""".format(
                    k, v_str
                )
            )
        return '<table style="margin: 0em"><tbody>{}</tbody></table>'.format("".join(rows))

    elif isinstance(record, list):
        record = list(filter(None, record))
        if not record:
            return None

        rows = []
        for item in record:
            rows.append("<li>{}</li>".format(to_html(item)))
        return '<ul style="text-align: left; margin-top: 0em; margin-bottom: 0em">{}</ul>'.format("".join(rows))

    else:
        return str(record)


def is_year_or_date(s: str) -> bool:
    """Matches dates in "yyyy-mm-dd" format or years in "yyyy" format."""
    date_pattern = r"^\d{4}-\d{2}-\d{2}$"
    year_pattern = r"^\d{4}$"

    if re.match(date_pattern, s) or re.match(year_pattern, s):
        return True
    else:
        return False


def _hash_dataclass(dataclass: Any) -> int:
    """Return unique hash for a dataclass. This is useful if you can't make your dataclasses
    frozen but still want to use operations such as `set` or `unique`."""
    fields = []
    for k, v in dataclass.__dict__.items():
        if is_dataclass(v):
            fields.append((k, hash(v)))
        elif isinstance(v, list):
            hashes = [_hash_dataclass(x) if is_dataclass(x) else hash(x) for x in v]
            fields.append((k, hash(tuple(hashes))))
        elif isinstance(v, dict):
            hashes = [(x, _hash_dataclass(y) if is_dataclass(y) else hash(y)) for x, y in v.items()]
            fields.append((k, hash(tuple(hashes))))
        else:
            fields.append((k, v))
    return hash(tuple(fields))


def _deepcopy_dataclass(dc) -> Any:
    """Create a deep copy of a dataclass. This is much faster than running copy.deepcopy."""
    dc = dataclasses.replace(dc)
    for k, v in dc.__dict__.items():
        if is_dataclass(v):
            setattr(dc, k, _deepcopy_dataclass(v))
        elif isinstance(v, list):
            lis = [_deepcopy_dataclass(x) if is_dataclass(x) else x for x in v]
            # make sure to preserve the type of the list if we subclass it
            if type(v) != list:  # noqa
                lis = type(v)(lis)
            setattr(dc, k, lis)
        elif isinstance(v, dict):
            setattr(dc, k, {x: _deepcopy_dataclass(y) if is_dataclass(y) else y for x, y in v.items()})
        else:
            pass
    return dc



#
#  catalog.py
#  owid-catalog-py
#

import heapq
import json
import os
import re
import tempfile
from pathlib import Path
from typing import Any, Dict, Iterable, Iterator, List, Optional, Union, cast
from urllib.parse import urlparse

import numpy as np
import numpy.typing as npt
import pandas as pd
import requests
import structlog

from . import s3_utils
from .datasets import CHANNEL, PREFERRED_FORMAT, SUPPORTED_FORMATS, Dataset, FileFormat
from .tables import Table

log = structlog.get_logger()

# increment this on breaking changes to require clients to update
OWID_CATALOG_VERSION = 3

# location of the default remote catalog
OWID_CATALOG_URI = "https://catalog.ourworldindata.org/"

# S3 location for private files
S3_OWID_URI = "s3://owid-catalog"

# global copy cached after first request
REMOTE_CATALOG: Optional["RemoteCatalog"] = None

# what formats should we for our index of available datasets?
INDEX_FORMATS: List[FileFormat] = ["feather"]


class CatalogMixin:
    """
    Abstract data catalog API, encapsulates finding and loading data.
    """

    channels: Iterable[CHANNEL]
    frame: "CatalogFrame"
    uri: str

    def find(
        self,
        table: Optional[str] = None,
        namespace: Optional[str] = None,
        version: Optional[str] = None,
        dataset: Optional[str] = None,
        channel: Optional[CHANNEL] = None,
    ) -> "CatalogFrame":
        criteria: npt.ArrayLike = np.ones(len(self.frame), dtype=bool)

        if table:
            criteria &= self.frame.table.str.contains(table)

        if namespace:
            criteria &= self.frame.namespace == namespace

        if version:
            criteria &= self.frame.version == version

        if dataset:
            criteria &= self.frame.dataset == dataset

        if channel:
            if channel not in self.channels:
                raise ValueError(
                    f"You need to add `{channel}` to channels in Catalog init (only `{self.channels}` are loaded now)"
                )
            criteria &= self.frame.channel == channel

        matches = self.frame[criteria]
        if "checksum" in matches.columns:
            matches = matches.drop(columns=["checksum"])

        return cast(CatalogFrame, matches)

    def find_one(self, *args: Optional[str], **kwargs: Optional[str]) -> Table:
        return self.find(*args, **kwargs).load()  # type: ignore

    def find_latest(
        self,
        *args: Optional[str],
        **kwargs: Optional[str],
    ) -> Table:
        frame = self.find(*args, **kwargs)  # type: ignore
        if frame.empty:
            raise ValueError("No matching table found")
        else:
            return cast(Table, frame.sort_values("version").iloc[-1].load())

    def __getitem__(self, path: str) -> Table:
        uri = "/".join([self.uri.rstrip("/"), path])
        for _format in SUPPORTED_FORMATS:
            try:
                return Table.read(f"{uri}.{_format}")
            except Exception:
                continue

        raise KeyError(f"no matching table found at: {uri}")


class LocalCatalog(CatalogMixin):
    """
    A data catalog that's on disk. On-disk catalogs do not need an index file, since
    you can simply walk the directory. However, they support a `reindex()` method
    which can create such an index.
    """

    uri: str

    def __init__(self, path: Union[str, Path], channels: Iterable[CHANNEL] = ("garden",)) -> None:
        self.uri = str(path)
        self.channels = channels
        if self._catalog_exists(channels):
            self.frame = CatalogFrame(self._read_channels(channels))
            self.frame._base_uri = self.path.as_posix() + "/"
        else:
            # could take a while to generate if there are many datasets
            self.reindex()

        # ensure the frame knows where to load data from

    @property
    def path(self) -> Path:
        return Path(self.uri)

    def _catalog_exists(self, channels: Iterable[CHANNEL]) -> bool:
        return all([self._catalog_channel_file(channel).exists() for channel in channels])

    def _catalog_channel_file(self, channel: CHANNEL, format: FileFormat = PREFERRED_FORMAT) -> Path:
        return self.path / f"catalog-{channel}.{format}"

    @property
    def _metadata_file(self) -> Path:
        return self.path / "catalog.meta.json"

    def _read_channels(self, channels: Iterable[CHANNEL]) -> pd.DataFrame:
        """
        Read selected channels from local path.
        """
        df = pd.concat([read_frame(self._catalog_channel_file(channel)) for channel in channels])
        df.dimensions = df.dimensions.map(lambda s: json.loads(s) if isinstance(s, str) else s)
        return df

    def iter_datasets(self, channel: CHANNEL, include: Optional[str] = None) -> Iterator[Dataset]:
        to_search = [self.path / channel]
        if not to_search[0].exists():
            return

        re_search = re.compile(include or "")

        while to_search:
            dir = heapq.heappop(to_search)
            if (dir / "index.json").exists() and re_search.search(str(dir)):
                yield Dataset(dir)
                continue

            for child in dir.iterdir():
                if child.is_dir():
                    heapq.heappush(to_search, child)

    def reindex(self, include: Optional[str] = None) -> None:
        """
        Walk the directory tree, generate a channel/namespace/version/dataset/table frame
        and save it to each of our index formats.
        """
        index = self._scan_for_datasets(include)

        if include:
            # we used regex to find datasets, so merge it with the original frame
            index = self._merge_index(self.frame, index)

        index._base_uri = self.path.as_posix() + "/"

        # convert int versions to strings
        index.version = index.version.astype(str)

        # make sure dimensions json is loaded
        index.dimensions = index.dimensions.map(lambda s: json.loads(s) if isinstance(s, str) else s)

        self._save_index(index)
        self.frame = index

    @staticmethod
    def _merge_index(frame: "CatalogFrame", update: "CatalogFrame") -> "CatalogFrame":
        """Merge two indexes."""
        return CatalogFrame(
            pd.concat(
                [update, frame.loc[~frame.path.isin(update.path)]],
                ignore_index=True,
            )
        )

    def _save_index(self, frame: "CatalogFrame") -> None:
        """
        Save all channels to disk in separate catalog files, and in each of our
        supported formats.
        """
        for channel in self.channels:
            channel_frame = frame.loc[frame.channel == channel].reset_index(drop=True)
            for format in INDEX_FORMATS:
                filename = self._catalog_channel_file(channel, format)
                save_frame(channel_frame, filename)

        # add a catalog version number that we can use to tell old clients to update
        self._save_metadata({"format_version": OWID_CATALOG_VERSION})

    def _scan_for_datasets(self, include: Optional[str] = None) -> "CatalogFrame":
        """Scan datasets. You can filter by `include` to get better performance."""
        frames = []
        log.info("reindex.start", channels=self.channels, include=include)
        for channel in self.channels:
            channel_frames = []
            for ds in self.iter_datasets(channel, include=include):
                channel_frames.append(ds.index(self.path))
            frames += channel_frames
            log.info(
                "reindex",
                channel=channel,
                datasets=len(channel_frames),
                include=include,
            )

        df = pd.concat(frames, ignore_index=True)

        keys = ["table", "dataset", "version", "namespace", "channel", "is_public"]
        columns = keys + [c for c in df.columns if c not in keys]

        df.sort_values(keys, inplace=True)  # type: ignore
        df = df.loc[:, columns]

        return CatalogFrame(df)

    def _save_metadata(self, contents: Dict[str, Any]) -> None:
        with open(self._metadata_file, "w") as ostream:
            json.dump(contents, ostream, indent=2)


class RemoteCatalog(CatalogMixin):
    uri: str

    def __init__(self, uri: str = OWID_CATALOG_URI, channels: Iterable[CHANNEL] = ("garden",)) -> None:
        self.uri = uri
        self.channels = channels
        self.metadata = self._read_metadata(self.uri + "catalog.meta.json")
        if self.metadata["format_version"] > OWID_CATALOG_VERSION:
            raise PackageUpdateRequired(
                f"library supports api version {OWID_CATALOG_VERSION}, "
                f'but the remote catalog has version {self.metadata["format_version"]} '
                "-- please update"
            )

        self.frame = CatalogFrame(self._read_channels(uri, channels))
        self.frame._base_uri = uri

    @property
    def datasets(self) -> pd.DataFrame:
        return self.frame[["namespace", "version", "dataset"]].drop_duplicates()

    @staticmethod
    def _read_metadata(uri: str) -> Dict[str, Any]:
        """
        Read the metadata JSON blob for this repo.
        """
        resp = requests.get(uri)
        resp.raise_for_status()
        return cast(Dict[str, Any], resp.json())

    @staticmethod
    def _read_channels(uri: str, channels: Iterable[CHANNEL]) -> pd.DataFrame:
        """
        Read selected channels from S3.
        """
        return pd.concat([read_frame(uri + f"catalog-{channel}.{PREFERRED_FORMAT}") for channel in channels])


class CatalogFrame(pd.DataFrame):
    """
    DataFrame helper, meant only for displaying catalog results.
    """

    _base_uri: Optional[str] = None

    _metadata = ["_base_uri"]

    @property
    def _constructor(self) -> type:
        return CatalogFrame

    @property
    def _constructor_sliced(self) -> Any:
        # ensure that when we pick a series we still have the URI
        def build(*args: Any, **kwargs: Any) -> Any:
            c = CatalogSeries(*args, **kwargs)
            c._base_uri = self._base_uri
            return c

        return build

    def load(self) -> Table:
        if len(self) == 1:
            return self.iloc[0].load()  # type: ignore
        elif len(self) == 0:
            raise ValueError("no tables found")
        else:
            raise ValueError(f"only one table can be loaded at once (tables found: {', '.join(self.table.tolist())})")

    @staticmethod
    def create_empty() -> "CatalogFrame":
        return CatalogFrame(
            {
                "namespace": [],
                "version": [],
                "table": [],
                "dimensions": [],
                "path": [],
                "format": [],
            }
        )


class CatalogSeries(pd.Series):
    """
    A row from the catalog representing a single dataset. We subclass Series
    in order to add a `load()` method onto it that will fetch and return a Table.
    """

    _metadata = ["_base_uri"]

    @property
    def _constructor(self) -> type:
        return CatalogSeries

    def load(self) -> Table:
        # determine what format to use for this table; old indexes gave one format,
        # new ones give multiple to choose from
        format = None
        if hasattr(self, "format"):
            # backwards compatibility with existing indexes
            format = self.format
        elif hasattr(self, "formats") and (self.formats is not None) and len(self.formats) > 0:
            format = PREFERRED_FORMAT if PREFERRED_FORMAT in self.formats else self.formats[0]

        if self.path and format and self._base_uri:
            with tempfile.TemporaryDirectory() as tmpdir:
                uri = self._base_uri + self.path + "." + format

                # download the data locally first if the file is private
                # keep backward compatibility
                if not getattr(self, "is_public", True):
                    uri = _download_private_file(uri, tmpdir)

                return Table.read(uri)

        raise ValueError("series is not a table spec")


def _load_remote_catalog(channels):
    global REMOTE_CATALOG

    # add channel if missing and reinit remote catalog
    if REMOTE_CATALOG and not (set(channels) <= set(REMOTE_CATALOG.channels)):
        REMOTE_CATALOG = RemoteCatalog(channels=list(set(REMOTE_CATALOG.channels) | set(channels)))

    if not REMOTE_CATALOG:
        REMOTE_CATALOG = RemoteCatalog(channels=channels)

    return REMOTE_CATALOG


def find(
    table: Optional[str] = None,
    namespace: Optional[str] = None,
    version: Optional[str] = None,
    dataset: Optional[str] = None,
    channels: Iterable[CHANNEL] = ("garden",),
) -> "CatalogFrame":
    REMOTE_CATALOG = _load_remote_catalog(channels=channels)

    return REMOTE_CATALOG.find(table=table, namespace=namespace, version=version, dataset=dataset)


def find_one(*args: Optional[str], **kwargs: Optional[str]) -> Table:
    return find(*args, **kwargs).load()  # type: ignore


def find_latest(
    table: Optional[str] = None,
    namespace: Optional[str] = None,
    dataset: Optional[str] = None,
    channels: Iterable[CHANNEL] = ("garden",),
    version: Optional[str] = None,
) -> Table:
    REMOTE_CATALOG = _load_remote_catalog(channels=channels)

    # If version is not specified, it will find the latest version given all other specifications.
    return REMOTE_CATALOG.find_latest(table=table, namespace=namespace, dataset=dataset, version=version)


def _download_private_file(uri: str, tmpdir: str) -> str:
    parsed = urlparse(uri)
    base, ext = os.path.splitext(parsed.path)
    s3_utils.download(
        S3_OWID_URI + base + ".meta.json",
        tmpdir + "/data.meta.json",
    )
    s3_utils.download(
        S3_OWID_URI + base + ext,
        tmpdir + "/data" + ext,
    )
    return tmpdir + "/data" + ext


class PackageUpdateRequired(Exception):
    pass


def read_frame(uri: Union[str, Path]) -> pd.DataFrame:
    if isinstance(uri, Path):
        uri = str(uri)

    if uri.endswith(".feather"):
        return cast(pd.DataFrame, pd.read_feather(uri))

    elif uri.endswith(".parquet"):
        return cast(pd.DataFrame, pd.read_parquet(uri))

    elif uri.endswith(".csv"):
        return pd.read_csv(uri)

    raise ValueError(f"could not detect format of uri: {uri}")


def save_frame(df: pd.DataFrame, path: Union[str, Path]) -> None:
    path = str(path)
    if path.endswith(".feather"):
        df.to_feather(path)

    elif path.endswith(".parquet"):
        df.to_parquet(path)

    elif path.endswith(".csv"):
        df.to_csv(path)

    else:
        raise ValueError(f"could not detect what format to write to: {path}")



"""There's a significant overlap with https://github.com/owid/walden/blob/master/owid/walden/owid_cache.py
It would make sense to move both into a shared module in the future or use some proper public library
for working with S3 that is compatible with DigitalOcean's Spaces.
"""
import logging
import os
from os import path
from typing import Any, Tuple
from urllib.parse import urlparse

from botocore.exceptions import ClientError

SPACES_ENDPOINT = "https://nyc3.digitaloceanspaces.com"
S3_BASE = "s3://walden.nyc3.digitaloceanspaces.com"
HTTPS_BASE = "https://walden.nyc3.digitaloceanspaces.com"
AWS_PROFILE = os.environ.get("AWS_PROFILE", "default")


def upload(filename: str, relative_path: str, public: bool = False) -> str:
    """
    Upload file to Walden.

    Args:
        local_path (str): Local path to file.
        walden_path (str): Path where to store the file in Walden.
        public (bool): Set to True to expose the file to the public (read only). Defaults to False.
    """
    dest_path = f"{S3_BASE}/{relative_path}"
    extra_args = {"ACL": "public-read"} if public else {}

    client = connect()
    try:
        client.upload_file(filename, "walden", relative_path, ExtraArgs=extra_args)
    except ClientError as e:
        logging.error(e)
        raise UploadError(e)

    logging.info("UPLOADED", f"{filename} -> {dest_path}")

    return f"{HTTPS_BASE}/{relative_path}"


def s3_bucket_key(url: str) -> Tuple[str, str]:
    """Get bucket and key from either s3:// URL or https:// URL."""
    parsed = urlparse(url)
    bucket = parsed.netloc
    key = parsed.path.lstrip("/")

    # strip region from bucket name in https scheme
    if parsed.scheme == "https":
        bucket = bucket.split(".")[0]

    return bucket, key


def download(s3_url: str, filename: str, quiet: bool = False) -> None:
    """Download the file at the S3 URL to the given local filename."""
    client = connect()

    bucket, key = s3_bucket_key(s3_url)

    try:
        client.download_file(bucket, key, filename)
    except ClientError as e:
        logging.error(e)
        raise UploadError(e)

    if not quiet:
        logging.info("DOWNLOADED", f"{s3_url} -> {filename}")


def connect() -> Any:
    "Return a connection to Walden's DigitalOcean space."
    import boto3

    check_for_default_profile()

    session = boto3.Session(profile_name=AWS_PROFILE)
    client = session.client(
        service_name="s3",
        endpoint_url=SPACES_ENDPOINT,
    )
    return client


def check_for_default_profile() -> None:
    filename = path.expanduser("~/.aws/config")
    if not path.exists(filename) or f"[{AWS_PROFILE}]" not in open(filename).read():
        raise MissingCredentialsError(
            f"""you must set up a config file at ~/.aws/config

it should look like:

[{AWS_PROFILE}]
aws_access_key_id = ...
aws_secret_access_key = ...
region = ...
"""
        )


class MissingCredentialsError(Exception):
    pass


class UploadError(Exception):
    pass



#
#  variables.py
#

import copy
import json
import os
from collections import defaultdict
from typing import Any, Dict, List, Literal, Optional, Union, cast, overload

import pandas as pd
import structlog
from pandas._typing import Scalar
from pandas.core.series import Series

from .meta import (
    PROCESSING_LEVELS,
    PROCESSING_LEVELS_ORDER,
    License,
    Origin,
    Source,
    VariableMeta,
    VariablePresentationMeta,
)
from .properties import metadata_property

log = structlog.get_logger()

SCHEMA = json.load(open(os.path.join(os.path.dirname(__file__), "schemas", "table.json")))
METADATA_FIELDS = list(SCHEMA["properties"])

# Defined operations.
OPERATION = Literal[
    "+",
    "-",
    "*",
    "/",
    "**",
    "//",
    "%",
    "fillna",
    "dropna",
    "load",
    "create",
    "save",
    "merge",
    "rename",
    "melt",
    "pivot",
    "concat",
    "sort",
    "pct_change",
]

# Environment variable such that, if True, the processing log will be updated, if False, the log will always be empty.
# If not defined, assume False.
PROCESSING_LOG = bool(os.getenv("PROCESSING_LOG", False))

# NOTE: The following issue seems to not be happening anymore. Consider deleting instances of UNNAMED_VARIABLE.
# When creating a new variable, might we need to pass a temporary name. For example, when doing tb["a"] + tb["b"]:
#  * If variable.name is None, a ValueError is raised.
#  * If variable.name = self.checked_name then the metadata of the first variable summed ("a") is modified.
#  * If variable.name is always a random string (that does not coincide with an existing variable) then
#    when replacing a variable (e.g. tb["a"] += 1) the original variable loses its metadata.
# For these reasons, we ensure that variable.name is always filled, even with a temporary name.
# In fact, if the new variable becomes a column in a table, its name gets overwritten by the column name (which is a
# nice feature). For example, when doing tb["c"] = tb["a"] + tb["b"], the variable name of "c" will be "c", even if we
# passed a temporary variable name. Therefore, this temporary name may be irrelevant in practice.
UNNAMED_VARIABLE = "**TEMPORARY UNNAMED VARIABLE**"


class Variable(pd.Series):
    _name: Optional[str] = None
    _fields: Dict[str, VariableMeta]

    def __init__(
        self,
        data: Any = None,
        index: Any = None,
        name: Optional[str] = None,
        _fields: Optional[Dict[str, VariableMeta]] = None,
        metadata: Optional[VariableMeta] = None,
        **kwargs: Any,
    ) -> None:
        if metadata:
            assert not _fields, "cannot pass both metadata and _fields"
            assert name or self.name, "cannot pass metadata without a name"
            _fields = {(name or self.name): metadata}  # type: ignore

        self._fields = _fields or defaultdict(VariableMeta)

        # silence warning
        if data is None and not kwargs.get("dtype"):
            kwargs["dtype"] = "object"

        super().__init__(data=data, index=index, name=name, **kwargs)

    @property
    def m(self) -> VariableMeta:
        """Metadata alias to save typing."""
        return self.metadata

    @property
    def name(self) -> Optional[str]:
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        # None name does not modify _fields, it is usually triggered on pandas operations
        if name is not None:
            # move metadata when you rename a field
            if self._name and self._name in self._fields:
                self._fields[name] = self._fields.pop(self._name)

            # make sure there is always a placeholder metadata object
            if name not in self._fields:
                self._fields[name] = VariableMeta()

        self._name = name

    @property
    def checked_name(self) -> str:
        if self.name is None:
            raise ValueError("variable must be named to have metadata")

        return self.name

    # which fields should pandas propagate on slicing, etc?
    _metadata = ["_fields", "_name"]

    @property
    def _constructor(self) -> type:
        return Variable

    @property
    def _constructor_expanddim(self) -> type:
        # XXX lazy circular import
        from . import tables

        return tables.Table

    @property
    def metadata(self) -> VariableMeta:
        vm = self._fields[self.checked_name]

        # pass this as a hidden attribute to the metadata object for display only
        vm._name = self.name  # type: ignore

        return vm

    @metadata.setter
    def metadata(self, meta: VariableMeta) -> None:
        self._fields[self.checked_name] = meta

    def astype(self, *args: Any, **kwargs: Any) -> "Variable":
        # To fix: https://github.com/owid/owid-catalog-py/issues/12
        v = super().astype(*args, **kwargs)
        v.name = self.name
        return cast(Variable, v)

    def _repr_html_(self):
        html = str(self)
        return """
             <h2 style="margin-bottom: 0em"><pre>{}</pre></h2>
             <p style="font-variant: small-caps; font-size: 1.5em; font-family: sans-serif; color: grey; margin-top: -0.2em; margin-bottom: 0.2em">variable</p>
             <pre>{}</pre>
        """.format(
            self.name, html
        )

    def __add__(self, other: Union[Scalar, Series, "Variable"]) -> "Variable":
        variable_name = self.name or UNNAMED_VARIABLE
        variable = Variable(super().__add__(other), name=variable_name)
        variable.metadata = combine_variables_metadata(variables=[self, other], operation="+", name=variable_name)
        return variable

    def __iadd__(self, other: Union[Scalar, Series, "Variable"]) -> "Variable":
        return self.__add__(other)

    def __sub__(self, other: Union[Scalar, Series, "Variable"]) -> "Variable":
        variable_name = self.name or UNNAMED_VARIABLE
        variable = Variable(super().__sub__(other), name=variable_name)
        variable.metadata = combine_variables_metadata(variables=[self, other], operation="-", name=variable_name)
        return variable

    def __isub__(self, other: Union[Scalar, Series, "Variable"]) -> "Variable":
        return self.__sub__(other)

    def __mul__(self, other: Union[Scalar, Series, "Variable"]) -> "Variable":
        variable_name = self.name or UNNAMED_VARIABLE
        variable = Variable(super().__mul__(other), name=variable_name)
        variable.metadata = combine_variables_metadata(variables=[self, other], operation="*", name=variable_name)
        return variable

    def __imul__(self, other: Union[Scalar, Series, "Variable"]) -> "Variable":
        return self.__mul__(other)

    def __truediv__(self, other: Union[Scalar, Series, "Variable"]) -> "Variable":
        variable_name = self.name or UNNAMED_VARIABLE
        variable = Variable(super().__truediv__(other), name=variable_name)
        variable.metadata = combine_variables_metadata(variables=[self, other], operation="/", name=variable_name)
        return variable

    def __itruediv__(self, other: Union[Scalar, Series, "Variable"]) -> "Variable":
        return self.__truediv__(other)

    def __floordiv__(self, other: Union[Scalar, Series, "Variable"]) -> "Variable":
        variable_name = self.name or UNNAMED_VARIABLE
        variable = Variable(super().__floordiv__(other), name=variable_name)
        variable.metadata = combine_variables_metadata(variables=[self, other], operation="//", name=variable_name)
        return variable

    def __ifloordiv__(self, other: Union[Scalar, Series, "Variable"]) -> "Variable":
        return self.__floordiv__(other)

    def __mod__(self, other: Union[Scalar, Series, "Variable"]) -> "Variable":
        variable_name = self.name or UNNAMED_VARIABLE
        variable = Variable(super().__mod__(other), name=variable_name)
        variable.metadata = combine_variables_metadata(variables=[self, other], operation="%", name=variable_name)
        return variable

    def __imod__(self, other: Union[Scalar, Series, "Variable"]) -> "Variable":
        return self.__mod__(other)

    def __pow__(self, other: Union[Scalar, Series, "Variable"]) -> "Variable":
        variable_name = self.name or UNNAMED_VARIABLE
        variable = Variable(super().__pow__(other), name=variable_name)
        variable.metadata = combine_variables_metadata(variables=[self, other], operation="**", name=variable_name)
        return variable

    def __ipow__(self, other: Union[Scalar, Series, "Variable"]) -> "Variable":
        return self.__pow__(other)

    def fillna(self, value=None, *args, **kwargs) -> "Variable":
        # NOTE: Argument "inplace" will modify the original variable's data, but not its metadata.
        #  But we should not use "inplace" anyway.
        if "inplace" in kwargs and kwargs["inplace"] is True:
            log.warning("Avoid using fillna(inplace=True), which may not handle metadata as expected.")
        variable_name = self.name or UNNAMED_VARIABLE
        variable = Variable(super().fillna(value, *args, **kwargs), name=variable_name)
        variable._fields = copy.deepcopy(self._fields)
        variable._fields[variable_name] = combine_variables_metadata(
            variables=[self, value], operation="fillna", name=variable_name
        )
        return variable

    def dropna(self, *args, **kwargs) -> "Variable":
        # NOTE: Argument "inplace" will modify the original variable's data, but not its metadata.
        #  But we should not use "inplace" anyway.
        if "inplace" in kwargs and kwargs["inplace"] is True:
            log.warning("Avoid using dropna(inplace=True), which may not handle metadata as expected.")
        variable_name = self.name or UNNAMED_VARIABLE
        variable = Variable(super().dropna(*args, **kwargs), name=variable_name)
        variable._fields = copy.deepcopy(self._fields)
        variable._fields[variable_name] = combine_variables_metadata(
            variables=[self], operation="dropna", name=variable_name
        )
        return variable

    def add(self, other: Union[Scalar, Series, "Variable"], *args, **kwargs) -> "Variable":
        if args or kwargs:
            raise NotImplementedError("This feature may exist in pandas, but not in owid.catalog.")
        return self.__add__(other=other)

    def sub(self, other: Union[Scalar, Series, "Variable"], *args, **kwargs) -> "Variable":
        if args or kwargs:
            raise NotImplementedError("This feature may exist in pandas, but not in owid.catalog.")
        return self.__sub__(other=other)

    def mul(self, other: Union[Scalar, Series, "Variable"], *args, **kwargs) -> "Variable":
        if args or kwargs:
            raise NotImplementedError("This feature may exist in pandas, but not in owid.catalog.")
        return self.__mul__(other=other)

    def truediv(self, other: Union[Scalar, Series, "Variable"], *args, **kwargs) -> "Variable":
        if args or kwargs:
            raise NotImplementedError("This feature may exist in pandas, but not in owid.catalog.")
        return self.__truediv__(other=other)

    def div(self, other: Union[Scalar, Series, "Variable"], *args, **kwargs) -> "Variable":
        return self.truediv(other=other, *args, **kwargs)

    def pct_change(self, *args, **kwargs) -> "Variable":
        variable_name = self.name or UNNAMED_VARIABLE
        variable = Variable(super().pct_change(*args, **kwargs), name=variable_name)
        variable._fields[variable_name] = combine_variables_metadata(
            variables=[self], operation="pct_change", name=variable_name
        )
        return variable

    def update_log(
        self,
        parents: List[Any],
        operation: str,
        variable_name: Optional[str] = None,
        comment: Optional[str] = None,
        inplace: bool = False,
    ) -> Optional["Variable"]:
        return update_log(
            variable=self,
            parents=parents,
            operation=operation,
            variable_name=variable_name,
            comment=comment,
            inplace=inplace,  # type: ignore
        )

    def amend_log(
        self,
        variable_name: Optional[str] = None,
        parents: Optional[List[Any]] = None,
        operation: Optional[str] = None,
        comment: Optional[str] = None,
        entry_num: int = -1,
        inplace: bool = False,
    ) -> Optional["Variable"]:
        return amend_log(
            variable=self,
            variable_name=variable_name,
            parents=parents,
            operation=operation,
            comment=comment,
            entry_num=entry_num,
            inplace=inplace,  # type: ignore
        )

    def copy_metadata(self, from_variable: "Variable", inplace: bool = False) -> Optional["Variable"]:
        return copy_metadata(to_variable=self, from_variable=from_variable, inplace=inplace)  # type: ignore

    def copy(self, deep: bool = True) -> "Variable":
        new_var = super().copy(deep=deep)
        if deep:
            new_var._fields = defaultdict(
                VariableMeta, {k: var_meta.copy(deep=deep) for k, var_meta in self._fields.items()}
            )
        return new_var


# dynamically add all metadata properties to the class
for k in VariableMeta.__dataclass_fields__:
    if hasattr(Variable, k):
        raise Exception(f'metadata field "{k}" would overwrite a Pandas built-in')

    setattr(Variable, k, metadata_property(k))


def _get_metadata_value_from_variables_if_all_identical(
    variables: List[Variable],
    field: str,
    warn_if_different: bool = False,
    operation: Optional[OPERATION] = None,
) -> Optional[Any]:
    if (operation == "/") and (getattr(variables[0].metadata, field) is None):
        # When dividing a variable by another, it only makes sense to keep the metadata values of the first variable.
        # For example, if we have energy (without description) and population (with a description), when calculating
        # energy per capita, the result shouldn't have the description of population. It should have no description.
        # Therefore, if the first variable has no metadata value, return None.
        return None

    # Get unique values from list, ignoring Nones.
    unique_values = set(
        [getattr(variable.metadata, field) for variable in variables if getattr(variable.metadata, field) is not None]
    )
    if len(unique_values) == 1:
        combined_value = unique_values.pop()
    else:
        combined_value = None
        if (len(unique_values) > 1) and warn_if_different:
            log.warning(f"Different values of '{field}' detected among variables: {unique_values}")

    return combined_value


def get_unique_sources_from_variables(variables: List[Variable]) -> List[Source]:
    # Make a list of all sources of all variables.
    sources = sum([variable.metadata.sources for variable in variables], [])

    return pd.unique(sources).tolist()


def get_unique_origins_from_variables(variables: List[Variable]) -> List[Origin]:
    # Make a list of all origins of all variables.
    origins = sum([variable.metadata.origins for variable in variables], [])

    # Get unique array of tuples of origin fields (respecting the order).
    return pd.unique(origins).tolist()


def get_unique_licenses_from_variables(variables: List[Variable]) -> List[License]:
    # Make a list of all licenses of all variables.
    licenses = sum([variable.metadata.licenses for variable in variables], [])

    return pd.unique(licenses).tolist()


def combine_variables_processing_logs(variables: List[Variable]) -> List[Dict[str, Any]]:
    # Make a list with all entries in the processing log of all variables.
    processing_log = sum(
        [
            variable.metadata.processing_log if variable.metadata.processing_log is not None else []
            for variable in variables
        ],
        [],
    )

    return processing_log


def _get_dict_from_list_if_all_identical(list_of_objects: List[Optional[Dict[str, Any]]]) -> Optional[Dict[str, Any]]:
    # The argument list_of_objects can contain dictionaries or None, or be empty.
    # If a list contains one dictionary (possibly repeated multiple times with identical content), return that
    # dictionary. Otherwise, if not all dictionaries are identical, return None.

    # List all dictionaries, ignoring Nones.
    defined_dicts = [d for d in list_of_objects if d is not None]

    if not defined_dicts:
        # If there are no dictionaries, return None.
        return None

    # Take the first dictionary as a reference.
    reference_dict = defined_dicts[0]

    # Return the first dictionary if all dictionaries are identical, otherwise return None.
    return reference_dict if all(d == reference_dict for d in defined_dicts) else None


def combine_variables_display(
    variables: List[Variable], operation: OPERATION, _field_name="display"
) -> Optional[Dict[str, Any]]:
    # Gather displays from all variables that are defined.
    list_of_displays = [getattr(variable.metadata, _field_name) for variable in variables]
    if operation == "/" and list_of_displays[0] is None:
        # When dividing a variable by another, it only makes sense to keep the display values of the first variable.
        # Therefore, if the first variables doesn't have a display, the resulting variable should have no display.
        return None
    else:
        return _get_dict_from_list_if_all_identical(list_of_objects=list_of_displays)


def combine_variables_presentation(
    variables: List[Variable], operation: OPERATION
) -> Optional[VariablePresentationMeta]:
    # Apply the same logic as for displays.
    return combine_variables_display(variables=variables, operation=operation, _field_name="presentation")  # type: ignore


def combine_variables_processing_level(variables: List[Variable]) -> Optional[PROCESSING_LEVELS]:
    # Gather processing levels from all variables that are defined.
    processing_levels = [
        variable.metadata.processing_level for variable in variables if variable.metadata.processing_level is not None
    ]

    if len(processing_levels) == 0:
        # If there are no processing levels, return None.
        return None

    # Ensure that all processing levels are known.
    unknown_processing_levels = set([level for level in processing_levels]) - set(PROCESSING_LEVELS_ORDER)
    assert len(unknown_processing_levels) == 0, f"Unknown processing levels: {unknown_processing_levels}"

    # If any of the variables has a processing level, take the highest level.
    maximum_level = max([PROCESSING_LEVELS_ORDER[level] for level in processing_levels])

    # Return the maximum level as a string.
    combined_processing_level = {value: key for key, value in PROCESSING_LEVELS_ORDER.items()}[maximum_level]

    return cast(PROCESSING_LEVELS, combined_processing_level)


def combine_variables_metadata(
    variables: List[Any], operation: OPERATION, name: str = UNNAMED_VARIABLE
) -> VariableMeta:
    # Initialise an empty metadata.
    metadata = VariableMeta()

    # Skip other objects passed in variables that may not contain metadata (e.g. a scalar),
    # and skip unnamed variables that cannot have metadata
    variables_only = [v for v in variables if hasattr(v, "name") and v.name and hasattr(v, "metadata")]

    # Combine each metadata field using the logic of the specified operation.
    metadata.title = _get_metadata_value_from_variables_if_all_identical(
        variables=variables_only, field="title", operation=operation
    )
    metadata.description = _get_metadata_value_from_variables_if_all_identical(
        variables=variables_only, field="description", operation=operation
    )
    metadata.description_short = _get_metadata_value_from_variables_if_all_identical(
        variables=variables_only, field="description_short", operation=operation
    )
    metadata.description_from_producer = _get_metadata_value_from_variables_if_all_identical(
        variables=variables_only, field="description_from_producer", operation=operation
    )
    metadata.unit = _get_metadata_value_from_variables_if_all_identical(
        variables=variables_only, field="unit", operation=operation, warn_if_different=True
    )
    metadata.short_unit = _get_metadata_value_from_variables_if_all_identical(
        variables=variables_only, field="short_unit", operation=operation, warn_if_different=True
    )
    metadata.sources = get_unique_sources_from_variables(variables=variables_only)
    metadata.origins = get_unique_origins_from_variables(variables=variables_only)
    metadata.licenses = get_unique_licenses_from_variables(variables=variables_only)
    metadata.processing_log = combine_variables_processing_logs(variables=variables_only)
    metadata.display = combine_variables_display(variables=variables_only, operation=operation)
    metadata.presentation = combine_variables_presentation(variables=variables_only, operation=operation)
    metadata.processing_level = combine_variables_processing_level(variables=variables_only)

    # List names of variables and scalars (or other objects passed in variables).
    variables_and_scalars_names = [
        variable.name if hasattr(variable, "name") else str(variable) for variable in variables
    ]
    metadata.processing_log = add_entry_to_processing_log(
        processing_log=metadata.processing_log,
        variable_name=name,
        parents=variables_and_scalars_names,
        operation=operation,
    )

    return metadata


def add_entry_to_processing_log(
    processing_log: List[Any],
    variable_name: str,
    parents: List[Any],
    operation: str,
    comment: Optional[str] = None,
) -> List[Any]:
    if not PROCESSING_LOG:
        # Avoid any processing and simply return the same input processing log.
        return processing_log

    # Consider using a deepcopy if any of the operations in this function alter mutable objects in processing_log.
    processing_log_updated = copy.deepcopy(processing_log)

    # TODO: Parents currently can be anything. Here we should ensure that they are strings. For example, we could
    # extract the name of the parent if it is a variable.

    # Define new log entry.
    log_new_entry = {"variable": variable_name, "parents": parents, "operation": operation}
    if comment is not None:
        log_new_entry["comment"] = comment

    # Add new entry to log.
    processing_log_updated += [log_new_entry]

    return processing_log_updated


@overload
def update_log(
    variable: Variable,
    parents: List[Any],
    operation: str,
    variable_name: Optional[str] = None,
    comment: Optional[str] = None,
    inplace: Literal[True] = True,
) -> None:
    ...


@overload
def update_log(
    variable: Variable,
    parents: List[Any],
    operation: str,
    variable_name: Optional[str] = None,
    comment: Optional[str] = None,
    inplace: Literal[False] = False,
) -> Variable:
    ...


def update_log(
    variable: Variable,
    parents: List[Any],
    operation: str,
    variable_name: Optional[str] = None,
    comment: Optional[str] = None,
    inplace: bool = False,
) -> Optional[Variable]:
    if not inplace:
        variable = copy.deepcopy(variable)

    if variable_name is None:
        # If a variable name is not specified, take it from the variable, or otherwise use UNNAMED_VARIABLE.
        variable_name = variable.name or UNNAMED_VARIABLE

    # Add new entry to the variable's processing log.
    variable.metadata.processing_log = add_entry_to_processing_log(
        processing_log=variable.metadata.processing_log,
        variable_name=variable_name,
        parents=parents,
        operation=operation,
        comment=comment,
    )

    if not inplace:
        return variable


def amend_entry_in_processing_log(
    processing_log: List[Dict[str, Any]],
    parents: Optional[List[Any]],
    operation: Optional[str],
    variable_name: Optional[str] = None,
    comment: Optional[str] = None,
    entry_num: Optional[int] = -1,
) -> List[Any]:
    if not PROCESSING_LOG:
        # Avoid any processing and simply return the same input processing log.
        return processing_log

    # Consider using a deepcopy if any of the operations in this function alter mutable objects in processing_log.
    processing_log_updated = copy.deepcopy(processing_log)

    fields = {"variable": variable_name, "parents": parents, "operation": operation, "comment": comment}
    for field, value in fields.items():
        if value:
            processing_log_updated[entry_num][field] = value  # type: ignore

    return processing_log_updated


@overload
def amend_log(
    variable: Variable,
    variable_name: Optional[str] = None,
    parents: Optional[List[Any]] = None,
    operation: Optional[str] = None,
    comment: Optional[str] = None,
    entry_num: int = -1,
    inplace: Literal[True] = True,
) -> None:
    ...


@overload
def amend_log(
    variable: Variable,
    variable_name: Optional[str] = None,
    parents: Optional[List[Any]] = None,
    operation: Optional[str] = None,
    comment: Optional[str] = None,
    entry_num: int = -1,
    inplace: Literal[False] = False,
) -> Variable:
    ...


def amend_log(
    variable: Variable,
    variable_name: Optional[str] = None,
    parents: Optional[List[Any]] = None,
    operation: Optional[str] = None,
    comment: Optional[str] = None,
    entry_num: int = -1,
    inplace: bool = False,
) -> Optional[Variable]:
    if not inplace:
        variable = variable.copy()

    variable.metadata.processing_log = amend_entry_in_processing_log(
        processing_log=variable.metadata.processing_log,
        parents=parents,
        operation=operation,
        variable_name=variable_name,
        comment=comment,
        entry_num=entry_num,
    )

    if not inplace:
        return variable


def update_variable_name(variable: Variable, name: str) -> None:
    """Update the name of an unnamed variable, as well as its processing log, to have a new name.

    Say you have a table tb with columns "a" and "b".
    If you create a new variable "c" as
    > variable_c = tb["a"] + tb["b"]
    the new variable will have UNNAMED_VARIABLE as name.
    Also, in the processing log, the variable will be cited as UNNAMED_VARIABLE.
    To change the variable name to something more meaningful (e.g. "c"), the current function can be used,
    > update_variable_name(variable=variable_c, name="c")
    This function will update the variable name (in place) and will replace all instances of UNNAMED_VARIABLE in the
    processing log to the new name.

    This function is already used when a variable is added to a table column, so that
    > tb["c"] = tb["a"] + tb["b"]
    will create a new variable with name "c" (which, in the processing log, will be referred to as "c").

    Parameters
    ----------
    variable : Variable
        Variable whose name is given by UNNAMED_VARIABLE.
    name : str
        New name to assign to the variable.
    """
    if hasattr(variable.metadata, "processing_log") and variable.metadata.processing_log is not None:
        variable.metadata.processing_log = json.loads(
            json.dumps(variable.metadata.processing_log).replace("**TEMPORARY UNNAMED VARIABLE**", name)
        )
    variable.name = name


@overload
def copy_metadata(from_variable: Variable, to_variable: Variable, inplace: Literal[False] = False) -> Variable:
    ...


@overload
def copy_metadata(from_variable: Variable, to_variable: Variable, inplace: Literal[True] = True) -> None:
    ...


def copy_metadata(from_variable: Variable, to_variable: Variable, inplace: bool = False) -> Optional[Variable]:
    if inplace:
        to_variable.metadata = from_variable.metadata.copy()
    else:
        new_variable = to_variable.copy()
        new_variable.metadata = from_variable.metadata.copy()
        return new_variable



import datetime as dt
import re
from pathlib import Path
from typing import Optional, TypeVar, Union, overload

import dynamic_yaml
import pytz
import yaml
from unidecode import unidecode

T = TypeVar("T")


def pruned_json(cls: T) -> T:
    orig = cls.to_dict  # type: ignore

    # only keep non-null public variables
    # make sure to call `to_dict` of nested objects as well
    cls.to_dict = lambda self, **kwargs: {  # type: ignore
        k: getattr(self, k).to_dict(**kwargs) if hasattr(getattr(self, k), "to_dict") else v
        for k, v in orig(self, **kwargs).items()
        if not k.startswith("_") and v not in [None, [], {}]
    }

    return cls


@overload
def underscore(name: str, validate: bool = True, camel_to_snake: bool = False) -> str:
    ...


@overload
def underscore(name: None, validate: bool = True, camel_to_snake: bool = False) -> None:
    ...


def underscore(name: Optional[str], validate: bool = True, camel_to_snake: bool = False) -> Optional[str]:
    """Convert arbitrary string to under_score. This was fine tuned on WDI bank column names.
    This function might evolve in the future, so make sure to have your use cases in tests
    or rather underscore your columns yourself.

    Parameters
    ----------
    name : str
        String to format.
    validate: bool, optional
        Whether to validate that the string is under_score. Defaults to True.
    camel_to_snake: bool, optional
        Whether to convert camelCase to snake_case. Defaults to False.

    Returns
    -------
    str:
        String using snake_case formatting.
    """
    if name is None:
        return None

    orig_name = name

    # camelCase to snake_case
    if camel_to_snake:
        name = _camel_to_snake(name)

    name = (
        name.replace(" ", "_")
        .replace("-", "_")
        .replace("—", "_")
        .replace("–", "_")
        .replace("‑", "_")
        .replace(",", "_")
        .replace(".", "_")
        .replace("\t", "_")
        .replace("?", "_")
        .replace('"', "")
        .replace("‘", "")
        .replace("\xa0", "_")
        .replace("’", "")
        .replace("`", "")
        .replace("−", "_")
        .replace("*", "_")
        .replace("“", "")
        .replace("”", "")
        .replace("#", "")
        .replace("^", "")
        .lower()
    )

    # replace special separators
    name = (
        name.replace("(", "__")
        .replace(")", "__")
        .replace(":", "__")
        .replace(";", "__")
        .replace("[", "__")
        .replace("]", "__")
    )

    # replace special symbols
    name = name.replace("/", "_")
    name = name.replace("|", "_")
    name = name.replace("=", "_")
    name = name.replace("%", "pct")
    name = name.replace("+", "plus")
    name = name.replace("us$", "usd")
    name = name.replace("$", "dollar")
    name = name.replace("&", "_and_")
    name = name.replace("<", "_lt_")
    name = name.replace(">", "_gt_")
    name = name.replace("≥", "_gte_")
    name = name.replace("≤", "_lte_")

    # replace quotes
    name = name.replace("'", "")

    # shrink triple underscore
    name = re.sub("__+", "__", name)

    # convert special characters to ASCII
    name = unidecode(name).lower()

    # strip leading and trailing underscores
    name = name.strip("_")

    # if the first letter is number, prefix it with underscore
    if re.match("^[0-9]", name):
        name = f"_{name}"

    # make sure it's under_score now, if not then raise NameError
    if validate:
        validate_underscore(name, f"`{orig_name}`")

    return name


def _camel_to_snake(name: str) -> str:
    """Convert string camelCase to snake_case.

    Reference: https://stackoverflow.com/a/1176023/5056599 CC BY-SA 4.0

    Example:
    >>> _camel_to_snake('camelCase')
    'camel_case'

    Parameters
    ----------
    name : str
        String using camelCase formatting.

    Returns
    -------
    str:
        String using snake_case formatting.
    """
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def underscore_table(t, *args, **kwargs):
    """Convert column and index names to underscore. Only for backwards compatibility.
    Using table.underscore() method is preferred."""
    return t.underscore(*args, **kwargs)


def validate_underscore(name: Optional[str], object_name: str = "Name") -> None:
    """Raise error if name is not snake_case."""
    if name is not None and not re.match("^[a-z_][a-z0-9_]*$", name):
        raise NameError(f"{object_name} must be snake_case. Change `{name}` to `{underscore(name, validate=False)}`")


def dynamic_yaml_load(path: Union[Path, str], params: dict = {}, return_dict=False) -> dict:
    with open(path) as istream:
        yd = dynamic_yaml.load(istream)

    yd.update(params)

    # additional parameters
    yd["TODAY"] = dt.datetime.now().astimezone(pytz.timezone("Europe/London")).strftime("%-d %B %Y")

    if return_dict:
        yd = yaml.safe_load(dynamic_yaml.dump(yd))

    return yd



"""Common operations performed on tables and variables.

"""
from .tables import (
    ExcelFile,
    concat,
    melt,
    merge,
    pivot,
    read_csv,
    read_excel,
    read_feather,
    read_from_dict,
    read_from_records,
    read_json,
)

__all__ = [
    "ExcelFile",
    "concat",
    "melt",
    "merge",
    "pivot",
    "read_csv",
    "read_feather",
    "read_excel",
    "read_from_dict",
    "read_from_records",
    "read_json",
]



__version__ = "0.1.0"

from . import processing, utils
from .catalogs import CHANNEL, LocalCatalog, RemoteCatalog, find, find_latest, find_one
from .datasets import Dataset
from .meta import (
    DatasetMeta,
    FaqLink,
    License,
    Origin,
    Source,
    TableMeta,
    VariableMeta,
    VariablePresentationMeta,
)
from .tables import Table
from .variables import Variable

__all__ = [
    "LocalCatalog",
    "RemoteCatalog",
    "find",
    "find_latest",
    "find_one",
    "Dataset",
    "Table",
    "Variable",
    "DatasetMeta",
    "TableMeta",
    "VariableMeta",
    "VariablePresentationMeta",
    "FaqLink",
    "Source",
    "Origin",
    "License",
    "utils",
    "CHANNEL",
    "processing",
]



#
#  datasets.py
#

import hashlib
import json
import shutil
import warnings
from dataclasses import dataclass
from glob import glob
from os import environ
from os.path import join
from pathlib import Path
from typing import Any, Iterator, List, Literal, Optional, Union

import numpy as np
import pandas as pd
import yaml

from . import tables, utils
from .meta import SOURCE_EXISTS_OPTIONS, DatasetMeta, TableMeta
from .properties import metadata_property

FileFormat = Literal["csv", "feather", "parquet"]

# the formats we can serialise and deserialise; in some cases they
# will be tried in this order if we don't specify one explicitly
SUPPORTED_FORMATS: List[FileFormat] = ["feather", "parquet", "csv"]

# the formats we generate by default
DEFAULT_FORMATS: List[FileFormat] = ["feather"]

# the format we use by default if we only need one
PREFERRED_FORMAT: FileFormat = "feather"

# sanity checks
assert set(DEFAULT_FORMATS).issubset(SUPPORTED_FORMATS)
assert PREFERRED_FORMAT in DEFAULT_FORMATS
assert SUPPORTED_FORMATS[0] == PREFERRED_FORMAT

# available channels in the catalog
CHANNEL = Literal["garden", "meadow", "grapher", "backport", "open_numbers", "examples", "explorers"]

# all pandas nullable dtypes
NULLABLE_DTYPES = [f"{sign}{typ}{size}" for typ in ("Int", "Float") for sign in ("", "U") for size in (8, 16, 32, 64)]


@dataclass
class Dataset:
    """
    A dataset is a folder full of data tables, with metadata available at `index.json`.
    """

    path: str
    metadata: "DatasetMeta"

    def __init__(self, path: Union[str, Path]) -> None:
        # for convenience, accept Path objects directly
        if isinstance(path, Path):
            self.path = path.as_posix()
        else:
            self.path = path

        self.metadata = DatasetMeta.load(self._index_file)

    @classmethod
    def create_empty(cls, path: Union[str, Path], metadata: Optional["DatasetMeta"] = None) -> "Dataset":
        path = Path(path)

        if path.is_dir():
            if not (path / "index.json").exists():
                raise Exception(f"refuse to overwrite non-dataset dir at: {path}")
            shutil.rmtree(path)

        path.mkdir(parents=True, exist_ok=True)

        metadata = metadata or DatasetMeta()

        index_file = path / "index.json"
        metadata.save(index_file)

        return Dataset(path.as_posix())

    def add(
        self,
        table: tables.Table,
        formats: List[FileFormat] = DEFAULT_FORMATS,
        repack: bool = True,
    ) -> None:
        """
        Add this table to the dataset by saving it in the dataset's folder. By default we
        save in multiple formats, but if you need a specific one (e.g. CSV for explorers)
        you can specify it.

        :param repack: if True, try to cast column types to the smallest possible type (e.g. float64 -> float32)
            to reduce binary file size. Consider using False when your dataframe is large and the repack is failing.
        """

        utils.validate_underscore(table.metadata.short_name, "Table's short_name")
        for col in list(table.columns) + list(table.index.names):
            utils.validate_underscore(col, "Variable's name")

        # non-unique index might be causing problems down the line and is typically a mistake
        if not table.index.is_unique:
            warnings.warn(
                f"Table `{table.metadata.short_name}` from dataset `{self.metadata.short_name}` has non-unique index"
            )

        if not table.primary_key:
            if "OWID_STRICT" in environ:
                raise PrimaryKeyMissing(
                    f"Table `{table.metadata.short_name}` does not have a primary_key -- please use t.set_index([col, ...], verify_integrity=True) to indicate dimensions before saving"
                )
            else:
                warnings.warn(
                    f"Table `{table.metadata.short_name}` does not have a primary_key -- please use t.set_index([col, ...], verify_integrity=True) to indicate dimensions before saving"
                )

        if not table.index.is_unique and "OWID_STRICT" in environ:
            [(k, dups)] = table.index.value_counts().head(1).to_dict().items()
            raise NonUniqueIndex(
                f"Table `{table.metadata.short_name}` has duplicate values in the index -- could you have made a mistake?\n\n"
                f"e.g. key {k} is repeated {dups} times in the index"
            )

        # check Float64 and Int64 columns for np.nan
        # see: https://github.com/owid/etl/issues/1334
        for col, dtype in table.dtypes.items():
            if dtype in NULLABLE_DTYPES:
                # pandas nullable types like Float64 have their own pd.NA instead of np.nan
                # make sure we don't use wrong nan, otherwise dropna and other methods won't work
                assert (
                    np.isnan(table[col]).sum() == 0
                ), f"Column `{col}` is using np.nan, but it should be using pd.NA because it has type {table[col].dtype}"

        # copy dataset metadata to the table
        table.metadata.dataset = self.metadata

        for format in formats:
            if format not in SUPPORTED_FORMATS:
                raise Exception(f"Format '{format}'' is not supported")

            table_filename = join(self.path, table.metadata.checked_name + f".{format}")
            table.to(table_filename, repack=repack)

    def __getitem__(self, name: str) -> tables.Table:
        stem = self.path / Path(name)

        for format in SUPPORTED_FORMATS:
            path = stem.with_suffix(f".{format}")
            if path.exists():
                t = tables.Table.read(path)
                # dataset metadata might have been updated, refresh it
                t.metadata.dataset = self.metadata
                return t

        raise KeyError(f"Table `{name}` not found, available tables: {', '.join(self.table_names)}")

    def __contains__(self, name: str) -> bool:
        return any((Path(self.path) / name).with_suffix(f".{format}").exists() for format in SUPPORTED_FORMATS)

    def save(self) -> None:
        assert self.metadata.short_name, "Missing dataset short_name"
        utils.validate_underscore(self.metadata.short_name, "Dataset's short_name")

        if not self.metadata.namespace:
            warnings.warn(f"Dataset {self.metadata.short_name} is missing namespace")

        # determine channel automatically from path
        # NOTE: shouldn't we force channel/namespace/version/short_name to be filled from path?
        # see https://github.com/owid/owid-catalog-py/pull/79#issue-1507959097 for discussion
        parts = str(self.path).split("/")
        if len(parts) >= 4:
            channel, _, _, _ = parts[-4:]
            if channel in CHANNEL.__args__:  # type: ignore
                self.metadata.channel = channel

        self.metadata.save(self._index_file)

        # Update the copy of this datasets metadata in every table in the set.
        for table_name in self.table_names:
            table = self[table_name]
            table.metadata.dataset = self.metadata
            table._save_metadata(join(self.path, table.metadata.checked_name + ".meta.json"))

    def update_metadata(self, metadata_path: Path, if_source_exists: SOURCE_EXISTS_OPTIONS = "replace") -> None:
        """
        Load YAML file with metadata from given path and update metadata of dataset and its tables.

        :param metadata_path: Path to *.meta.yml file with metadata. Check out other metadata files
            for examples, this function doesn't do schema validation
        :param if_source_exists: What to do if source already exists in metadata. Possible values:
            - "replace" (default): replace existing source with new one
            - "append": append new source to existing ones
            - "fail": raise an exception if source already exists
        """
        self.metadata.update_from_yaml(metadata_path, if_source_exists=if_source_exists)

        with open(metadata_path) as istream:
            metadata = yaml.safe_load(istream)
            for table_name in metadata.get("tables", {}).keys():
                table = self[table_name]
                table.update_metadata_from_yaml(metadata_path, table_name)
                table._save_metadata(join(self.path, table.metadata.checked_name + ".meta.json"))

    def index(self, catalog_path: Path = Path("/")) -> pd.DataFrame:
        """
        Return a DataFrame describing the contents of this dataset, one row per table.
        """
        base = {
            "namespace": self.metadata.namespace,
            "dataset": self.metadata.short_name,
            "version": self.metadata.version,
            "checksum": self.checksum(),
            "is_public": self.metadata.is_public,
        }
        rows = []
        for metadata_file in self._metadata_files:
            with open(metadata_file) as istream:
                metadata = TableMeta.from_dict(json.load(istream))

            row = base.copy()

            assert metadata.short_name
            row["table"] = metadata.short_name

            row["dimensions"] = json.dumps(metadata.primary_key)

            table_path = Path(self.path) / metadata.short_name
            relative_path = table_path.relative_to(catalog_path)
            row["path"] = relative_path.as_posix()
            row["channel"] = relative_path.parts[0]

            row["formats"] = [f for f in SUPPORTED_FORMATS if table_path.with_suffix(f".{f}").exists()]  # type: ignore

            rows.append(row)

        return pd.DataFrame.from_records(rows)

    @property
    def _index_file(self) -> str:
        return join(self.path, "index.json")

    def __bool__(self) -> bool:
        return True

    def __len__(self) -> int:
        return len(self.table_names)

    def __iter__(self) -> Iterator[tables.Table]:
        for name in self.table_names:
            yield self[name]

    @property
    def _data_files(self) -> List[str]:
        files = []
        for format in SUPPORTED_FORMATS:
            pattern = join(self.path, f"*.{format}")
            files.extend(glob(pattern))

        return sorted(files)

    @property
    def table_names(self) -> List[str]:
        return sorted(set(Path(f).stem for f in self._data_files))

    @property
    def _metadata_files(self) -> List[str]:
        return sorted(glob(join(self.path, "*.meta.json")))

    def checksum(self) -> str:
        "Return a MD5 checksum of all data and metadata in the dataset."
        _hash = hashlib.md5()
        _hash.update(checksum_file(self._index_file).digest())

        for data_file in self._data_files:
            _hash.update(checksum_file(data_file).digest())

            metadata_file = Path(data_file).with_suffix(".meta.json").as_posix()
            _hash.update(checksum_file(metadata_file).digest())

        return _hash.hexdigest()


for k in DatasetMeta.__dataclass_fields__:
    if hasattr(Dataset, k):
        raise Exception(f'metadata field "{k}" would overwrite a Dataset built-in')

    setattr(Dataset, k, metadata_property(k))


def checksum_file(filename: str) -> Any:
    "Return the MD5 checksum of a given file."
    chunk_size = 2**20  # 1MB
    checksum = hashlib.md5()
    with open(filename, "rb") as istream:
        chunk = istream.read(chunk_size)
        while chunk:
            checksum.update(chunk)
            chunk = istream.read(chunk_size)

    return checksum


class PrimaryKeyMissing(Exception):
    pass


class NonUniqueIndex(Exception):
    pass
